"""
Provide utility functions for extracting type annotations.

Functions
---------
get_annotations(object_instance: Any, root: type) -> dict[str, type]
    Extract all type annotations from an object instance up to a
    specified root class.
"""

from typing import Any


def get_annotations(object_instance: Any, root: type) -> dict[str, type]:
    """
    Extract type annotations from an object's class.

    Extract all type annotations from an object instance class up to,
    but not including, a specified root class. The annotations are
    collected from the class hierarchy in the order of method
    resolution.

    Parameters
    ----------
    object_instance : Any
        An instance of a class.
    root : type
        The root class from which to extract annotations.

    Returns
    -------
    dict[str, type]
        A dictionary of attribute names and their corresponding types.
    """
    all_attributes: dict[str, type] = {}
    mro_classes = object_instance.__class__.__mro__
    for cls in mro_classes:
        if not issubclass(cls, root) or cls is root:
            continue
        all_attributes |= {
            key: value
            for key, value in cls.__annotations__.items()
            if key not in all_attributes
        }
    return all_attributes



"""
Export the class for the downloader object.

The downloader object is responsible for downloading files from a
specified location.

Classes
-------
Downloader
    Abstract a downloader object.
"""

from .downloader import Downloader

__all__ = ["Downloader"]



"""
Define the base class for the downloader object.

The downloader object is responsible for downloading files from a
specified location using a datasource and product locator.

Classes
-------
DownloaderBase
    Represent a downloader object.
"""

import os
from dataclasses import dataclass
from datetime import datetime, timedelta

from ..dataset import ProductLocator
from ..datasource import Datasource
from ..utils import FileRepository
from .constants import (
    ISO_TIMESTAMP_FORMAT,
    TIME_TOLERANCE_DEFAULT,
    TIME_TOLERANCE_MAX,
    TIME_TOLERANCE_MIN,
    TIME_TOLERANCE_SINGLE,
    DownloadStatus,
)


@dataclass(eq=False, frozen=True)
class DownloaderBase:
    """
    Represent a downloader object.

    Methods
    -------
    download_files(start_time: str, end_time: str = "")
        Download files from the datasource into the local repository.
    get_files(start_time: str, end_time: str = "")
        Load a list of files from the datasource or local repository.
    list_files(start_time: str, end_time: str = "")
        List the files that can be retrieved from the datasource.

    Attributes
    ----------
    datasource : Datasource
        A reference to the datasource object.
    locator : ProductLocator
        A reference to the dataset's product locator object.
    repository : FileRepository
        A reference to the local repository object.
    date_format : str
        The date format specification. The default is the full ISO 8601
        timestamp format.
    time_tolerance : int
        The time tolerance in seconds. The default is 60 seconds.
    show_progress : bool
        A flag to show the progress of the download. The default is
        True, to display the progress.
    """

    datasource: Datasource
    locator: ProductLocator
    repository: FileRepository
    date_format: str = ISO_TIMESTAMP_FORMAT
    time_tolerance: int = TIME_TOLERANCE_DEFAULT
    show_progress: bool = True

    def __post_init__(self) -> None:
        """
        Validate the downloader object.

        Raises
        ------
        ValueError
            If the `time_tolerance` is less than the minimum allowed
            value or greater than the maximum allowed value.
        """
        if self.time_tolerance < TIME_TOLERANCE_MIN:
            raise ValueError(
                "time_tolerance must be greater than or equal "
                f"to {TIME_TOLERANCE_MIN} seconds"
            )
        if self.time_tolerance > TIME_TOLERANCE_MAX:
            raise ValueError(
                "time_tolerance must be less than or equal "
                f"to {TIME_TOLERANCE_MAX} seconds"
            )

    def download_files(self, *, start: str, end: str = "") -> list[str]:
        """
        Download files from the datasource into the local repository.

        Download the files that match the timestamps between `start` and
        `end` times, inclusive, from the datasource and save them in the
        local repository. The list is filtered by the timestamps of the
        files; only files in the requested range are downloaded. If the
        file is already in the local repository, it is not downloaded
        again.

        Note that a `start` date must be always provided. If an `end`
        date is not given, it is set equal to `start_time`. An offset of
        `time_tolerance` seconds (60 by default) is subtracted from the
        initial datetime and added to the final datetime to account for
        possible differences in the files' timestamps.

        Parameters
        ----------
        start : str
            The start time in the format specified by the `date_format`
            attribute.
        end : str
            The end time in the format specified by the `date_format`
            attribute. The default is "", in which case `end_time` is
            set equal to `start_time`.

        Returns
        -------
        list[str]
            A list of file path and names with respect to the local
            repository root directory.

        Notes
        -----
        `ValueError` is raised if `start_time` is not provided or if the
        provided timestamp strings and format can't be parsed (which is,
        indeed, a bug!).

        The framework may raise `RuntimeError` if the file cannot be
        retrieved, e.g. if the file does not exist in the datasource or
        an internal error occurred.

        `FileExistsError` might be raised if the file already exists in
        the repository.
        """
        files_in_range = self._get_file_list(start, end)

        self._retrieve_files(files_in_range)

        return files_in_range

    def get_files(self, *, file_paths: list[str]) -> None:
        """
        Load a list of files from the datasource or local repository.

        Load the files in the `file_paths` list from the datasource or
        local repository. If the file is not in the local repository, it
        is retrieved from the datasource and saved in the local
        repository.

        Parameters
        ----------
        file_paths : list[str]
            A list with the file paths.

        Notes
        -----
        The framework may raise `RuntimeError` if the file cannot be
        retrieved, e.g. if the file does not exist in the datasource or
        an internal error occurred.

        `FileExistsError` might be raised if the file already exists in
        the repository.
        """
        self._retrieve_files(file_paths)

    def list_files(self, *, start: str, end: str = "") -> list[str]:
        """
        List the files that can be retrieved from the datasource.

        List the files that match the timestamps between `start` and
        `end` times, inclusive, from the datasource or local repository.
        The list is filtered by the timestamps of the files; only files
        in the requested range are returned. This list can be filtered
        using custom criteria then passed to the `get_files()` method.

        Note that a `start` date must be always provided. If an `end`
        date is not given, it is set equal to `start_time`. An offset of
        `time_tolerance` seconds (60 by default) is subtracted from the
        initial datetime and added to the final datetime to account for
        possible differences in the files' timestamps.

        Parameters
        ----------
        start : str
            The start time in the format specified by the `date_format`
            attribute.
        end : str
            The end time in the format specified by the `date_format`
            attribute. The default is "", in which case `end_time` is
            set equal to `start_time`.

        Returns
        -------
        list[str]
            A list of file path and names with respect to the local
            repository root directory.

        Notes
        -----
        `ValueError` is raised if `start_time` is not provided or if the
        provided timestamp strings and format can't be parsed (which is,
        indeed, a bug!).
        """
        return self._get_file_list(start, end)

    def _add_item(self, file_path: str, file: bytes) -> None:
        if self._has_item(file_path):
            raise FileExistsError(f"File '{file_path}' already in repository.")

        self.repository.save_file(file, file_path)

    def _dowload_file(self, file_path: str) -> DownloadStatus:
        if self._has_item(file_path):
            return DownloadStatus.ALREADY

        content = self.datasource.download_file(file_path)

        self._add_item(file_path, content)

        return DownloadStatus.SUCCESS

    def _filter_directory_content(
        self, datetime_ini: datetime, datetime_fin: datetime, files: list[str]
    ) -> list[str]:
        files_in_range: list[str] = []

        for file in files:
            basename: str = os.path.basename(file)
            if not self.locator.match(basename):
                continue

            ct = self.locator.get_datetime(file)

            if datetime_ini <= ct <= datetime_fin:
                files_in_range.append(file)

        return files_in_range

    def _get_datetimes(
        self, start_time: str, end_time: str
    ) -> tuple[datetime, datetime]:
        if not start_time:
            raise ValueError("start_time must be provided")

        time_tolerance = self.time_tolerance

        datetime_ini = datetime.strptime(start_time, self.date_format)

        if end_time:
            datetime_fin = datetime.strptime(end_time, self.date_format)
        else:
            time_tolerance = max(time_tolerance, TIME_TOLERANCE_SINGLE)
            datetime_fin = datetime_ini

        # Sometimes the files have a date and time with some seconds
        # sooner or later to the user required times. To overcome this
        # issue, we subtract and add 60 seconds to the initial and final
        # datetimes, respectively.  For single instant downloading, a
        # special value `TIME_TOLERANCE_SINGLE` will be used as the
        # minimum tolerance in order to seek for the latest dataset for
        # the given instant.
        datetime_ini -= timedelta(seconds=time_tolerance)
        datetime_fin += timedelta(seconds=time_tolerance)

        return datetime_ini, datetime_fin

    def _get_file_list(self, start_time: str, end_time: str = "") -> list[str]:
        if self.show_progress:
            print("Retrieving available file list")

        datetime_ini, datetime_fin = self._get_datetimes(start_time, end_time)

        paths = self.locator.get_paths(datetime_ini, datetime_fin)

        files = self._retrieve_directory_content(paths)

        return self._filter_directory_content(
            datetime_ini, datetime_fin, files
        )

    def _has_item(self, file_path: str) -> bool:
        return self.repository.is_file(file_path)

    def _retrieve_directory_content(self, paths: list[str]) -> list[str]:
        files: list[str] = []

        for path in paths:
            files.extend(self.datasource.list_files(path))

        return files

    def _retrieve_files(self, file_paths: list[str]) -> None:
        num_files = len(file_paths)
        num_len = len(f"{num_files}")
        padding = " ".rjust(2 * num_len + 2)

        if self.show_progress:
            print("Downloading files:")

        for i, file_path in enumerate(file_paths):
            if self.show_progress:
                num_item = f"{i + 1}".rjust(num_len)
                print(f"{num_item}/{num_files} {file_path}")

            result = self._dowload_file(file_path)

            if self.show_progress:
                if result == DownloadStatus.SUCCESS:
                    print(f"{padding}... downloaded succesfully")
                else:
                    print(f"{padding}... already downloaded")



"""
Provide a locator for the GridSat-B1 imagery dataset products.

Classes
-------
GridSatProductLocatorB1
    Represent the GridSat-B1 imagery dataset product locator.
"""

from datetime import datetime

from .constants import (
    B1_FILE_DATE_FORMAT,
    B1_FILE_DATE_PATTERN,
    B1_FILE_PREFIX,
    B1_PATH_DATE_FORMAT,
    B1_PATH_PREFIX,
    B1_PRODUCT_LATEST_VERSION,
    B1_PRODUCT_NAME,
)
from .locator import GridSatProductLocator


class GridSatProductLocatorB1(GridSatProductLocator):
    """
    Represent the GridSat-B1 imagery dataset product locator.

    This class implements the `GridSatProductLocator` abstract class for
    for the GridSat-B1 (Geostationary IR Channel Brightness Temperature
    - GridSat B1) imagery dataset product locator.

    Instances of this class are responsible for generating a list of
    folder paths based on the dataset's directory structure and naming
    conventions, product details, and a specified date range. The
    generated paths cover each year within the date range. Paths to the
    folders containing the initial and final dates are included in the
    list.

    Instances of this class are also responsible for verifying if a
    given filename matches the product filename pattern based on the
    dataset's naming conventions and product specifications, and for
    extracting the corresponding `datetime` information from the
    product's filename.

    The data in the GridSat-B1 dataset products comes from different
    sources and only a global view of the Earth is available, thus,
    the origin's name and scene identifier are not reflected in the
    product's path or filename, but the product's version is. The
    product's filename pattern is as follows:

    'GRIDSAT-B1.<yyyy>.<mm>.<dd>.<HH>.<version>.nc';

    where `<yyyy>` is the gregorian year number; `<mm>`, `<dd>` and
    `<HH>` are the month, day and hour, respectively, using two digits
    padded with zeros; and `<version>` is the product's version (e.g.
    'v02r01').

    The dataset is available from various services; see the Notes below
    for a list of supported sources. For Amazon Web Services (AWS) and
    Google Cloud Platform (GCP), the product's file path pattern is as
    follows:

    '<scheme>://<net-location>/data/<yyyy>',

    where `<scheme>` identifies the protocol's scheme (e.g. 's3', 'gs'),
    `<net-location>` is the hostname (e.g. 'noaa-cdr-gridsat-b1-pds' for
    's3', and 'noaa-cdr-gridsat-b1' for 'gs'), and  `<yyyy>` is the
    gregorian year number.

    The files are also hosted on the servers of the National Centers for
    Environmental Information (NCEI), the product's file path pattern is
    as follows:

    'https://www.ncei.noaa.gov/data/<path>/access/<yyyy>';

    where `<path>` is the project's data files folder name, i.e.
    'geostationary-ir-channel-brightness-temperature-gridsat-b1',
    and `<yyyy>` is the gregorian year number.

    Input is 3-hourly data from the International Satellite Cloud
    Climatology Project (ISCCP) with gridded 0.07°x0.07° spatial
    resolution that spans from 1980 to the present. Three total
    channels are available:

    - IR: CDR-quality infrared window (IRWIN) channel (near 11 μm);
    - WV: Infrared water vapor (IRWVP) channel (near 6.7 μm);
    - VIS: Visible channel (near 0.6 μm).

    For more information visit the following link and links therein:
    https://www.ncei.noaa.gov/products/gridded-geostationary-brightness-temperature

    Notes
    -----
    Only the AWS datasource is supported currently.

    Methods
    -------
    get_base_url(datasource: str)
        Get the base URL for the GridSat-B1 imagery dataset's products.
    next_time(current_time: datetime)
        Get the next time interval. GridSat-B1 dataset organises the
        data by year.
    normalize_times(datetime_ini: datetime, datetime_fin: datetime)
        Normalise the initial and final datetimes.
    truncate_to_year(time: datetime)
        Truncate the `datetime` to the current year.

    Caution
    -------
    Members of this class not defined by the `ProductLocator` interface
    are helper methods and can be considered as implementation details,
    even though they are defined as part of the public API. In future
    releases, these methods may be moved to a private scope, suffer
    name changes, or be removed altogether.
    """  # noqa: E501

    # Base URLs for the available datasources of the GridSat-B1 dataset
    # Products:
    AVAILABLE_DATASOURCES: dict[str, str] = {
        "AWS": "s3://noaa-cdr-gridsat-b1-pds/data/",
        "GCP": "gs://noaa-cdr-gridsat-b1/data/",
        "HTTP": "https://www.ncei.noaa.gov/data/geostationary-ir-"
        "channel-brightness-temperature-gridsat-b1/access/",
    }

    # Supported datasources of the GridSat-B1 dataset Products:
    SUPPORTED_DATASOURCES: set[str] = {"AWS", "HTTP"}

    # Supported versions of the GridSat-B1 dataset Products:
    SUPPORTED_VERSIONS: set[str] = {"v02r01"}

    def __init__(
        self, versions: str | list[str] = B1_PRODUCT_LATEST_VERSION
    ) -> None:
        """
        Initialise a GridSat-B1 imagery dataset product locator.

        Constructs a new GridSat-B1 imagery dataset product locator
        object.

        Parameters
        ----------
        versions : str | list[str], optional
            The version of the GridSat-B1 product; e.g., "v02r01". The
            version may be a single version or a list of versions. Only
            the latest version is available in the public repository.
            The default is the latest version ("v02r01").

        Raises
        ------
        ValueError
            If the provided version is not supported or unavailable.
        """
        if isinstance(versions, str):
            versions = [versions]

        if unsupported_version := set(versions) - set(self.SUPPORTED_VERSIONS):
            supported_versions: list[str] = sorted(self.SUPPORTED_VERSIONS)
            raise ValueError(
                f"Unsupported version: {sorted(unsupported_version)}. "
                f"Supported versions: {supported_versions}"
            )

        super().__init__(
            name=B1_PRODUCT_NAME,
            origins=[],
            versions=versions,
            file_date_format=B1_FILE_DATE_FORMAT,
            file_date_pattern=B1_FILE_DATE_PATTERN,
            file_prefix=B1_FILE_PREFIX,
            path_date_format=B1_PATH_DATE_FORMAT,
            path_prefix=B1_PATH_PREFIX,
        )

    def get_base_url(self, datasource: str) -> tuple[str, ...]:
        """
        Get the base URL for the GridSat-B1 imagery dataset's products.

        This method returns the base URL for the GridSat-B1 imagery
        dataset's products. The base URL is used to construct the full
        URL to the dataset's product files.

        Parameters
        ----------
        datasource : str
            The datasource identifier. This parameter is used to
            determine the base URL for the dataset's products. The
            available datasources are 'AWS', 'GCP', and 'NOAA'. Only
            the AWS datasource is supported so far.

        Returns
        -------
        tuple[str, ...]
            The base URL for the GridSat-B1 imagery dataset's products
            based on the requested datasource identifier.

        Raises
        ------
        ValueError
            If the requested datasource is not supported or unavailable.
        """
        if datasource not in self.SUPPORTED_DATASOURCES:
            supported_datasources: list[str] = sorted(
                self.SUPPORTED_DATASOURCES
            )
            raise ValueError(
                f"Unsupported datasource: '{datasource}'. "
                f"Supported datasources: {supported_datasources}"
            )

        return (self.AVAILABLE_DATASOURCES[datasource], "")

    def next_time(self, current_time: datetime) -> datetime:
        """
        Get the next time interval.

        Get the next time interval based on the current time interval.

        Parameters
        ----------
        current_time : datetime
            The current time interval.

        Returns
        -------
        datetime
            The next time interval.
        """
        next_year: int = current_time.year + 1

        return current_time.replace(year=next_year)

    def normalize_times(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> tuple[datetime, datetime]:
        """
        Normalise the initial and final datetimes.

        Normalise the initial and final datetimes to the nearest
        time interval based on the dataset. The initial datetime is
        normalised to the start of the time interval and the final
        datetime is normalised to the end of the time interval.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime.

        datetime_fin : datetime
            The final datetime.

        Returns
        -------
        tuple[datetime, datetime]
            A tuple containing the normalised initial and final
            datetimes.
        """
        start_time: datetime = self.truncate_to_year(datetime_ini)
        end_time: datetime = self.truncate_to_year(datetime_fin)

        return start_time, end_time

    @staticmethod
    def truncate_to_year(time: datetime) -> datetime:
        """
        Truncate the `datetime` to the current year.

        The `datetime` is truncated to the beginning of the current
        year.

        Parameters
        ----------
        time : datetime
            The `datetime` to be truncated to the current year.

        Returns
        -------
        datetime
            The `datetime` truncated to the current year.
        """
        return time.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0
        )



"""
Provide a base class for data records extracted from netCDF datasets.

Defines the DataFragment class, which represents a data record extracted
from a netCDF dataset. The class provides methods for initializing
attributes from the dataset, validating preconditions and
postconditions, and preventing accidental modification of attributes.

Classes
-------
DataFragment
    Represents a data record extracted from a netCDF dataset.
"""

from inspect import currentframe, getmodule
from typing import Any, NoReturn, TypeVar, cast

from netCDF4 import Dataset  # pylint: disable=no-name-in-module

from .annotations import get_annotations
from .class_help import HasStrHelp, help_str
from .fields import ClassField, ViewField
from .hinting import get_annotated, get_typehint
from .validation import validate_type


class DatasetView(HasStrHelp):
    """
    Represent a data fragment extracted from a netCDF dataset.

    Attributes are dynamically set based on the annotations and field
    specifications defined in the class and the corresponding records
    present in the provided netCDF dataset.
    """

    def __init__(self, record: Dataset, **kwargs: Any) -> None:
        """
        Initialize the DataFragment object.

        Use annotations and field specification to initialize the
        DataFragment object with the provided netCDF dataset.

        Parameters
        ----------
        record : Dataset
            The netCDF dataset containing the data to be extracted.
        """
        self._validate_preconditions()
        self._init_class_attributes(record)
        self._init_record_attributes(record)
        self._init_field_attributes(record)
        self._perform_post_init_setup(record, **kwargs)
        self._validate_postconditions()

    def __post_init__(self, record: Dataset, **kwargs: Any) -> None:
        """
        Perform post-initialization setup.

        This method is called after the initialization of the DataFragment
        object. It can be overridden by subclasses to perform additional
        setup tasks.
        """

    def _perform_post_init_setup(self, record: Dataset, **kwargs: Any) -> None:
        self.__post_init__(record, **kwargs)

    def __delattr__(self, name: str) -> NoReturn:
        # Prevents deletion of attributes.
        raise AttributeError(f"Cannot delete attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        # Prevents setting of attributes outside of the __post_init__
        # method. Note tha this can always be bypassed by calling
        # object.__setattr__, but the goal is to prevent accidental
        # assignment of attributes to the instance not to guarantee
        # immutability.
        current = currentframe()
        frame = current.f_back if current else None
        while frame:
            if (
                frame.f_code.co_name == "__post_init__"
                and frame.f_locals.get("self") is self
            ):
                object.__setattr__(self, name, value)
                return
            frame = frame.f_back
        raise AttributeError(f"Cannot set attribute '{name}'")

    def __str__(self) -> str:
        # Returns a string representation of the DataFragment object.
        return help_str(self, DatasetView)

    def _init_class_attributes(self, record: Dataset) -> None:
        # Initializes class attributes from the provided dataset. Copy
        # class attributes to the instance provided they are not field
        # attributes.
        annotations = get_annotations(self, DatasetView)
        for name, annotation in annotations.items():
            if hasattr(self, name):
                value = getattr(self, name)
                if isinstance(value, ViewField):
                    continue
                if isinstance(value, ClassField):
                    value = value.value
                elif hasattr(record, name):
                    # This is a name clash, rename the attribute or
                    # explicitly make it a class field (ClassField).
                    atype = get_annotated(annotation)
                    raise AttributeError(
                        f"Attribute '{name}: {atype}' "
                        "collides with record attribute of the same name"
                    )
                object.__setattr__(self, name, value)

    def _init_record_attributes(self, record: Dataset) -> None:
        # Initializes record attributes from the provided dataset. Copy
        # record attributes to the instance.
        annotations = get_annotations(self, DatasetView)
        for name, annotation in annotations.items():
            if hasattr(self, name):
                continue
            if not hasattr(record, name):
                atype = get_annotated(annotation)
                raise AttributeError(
                    f"Attribute '{name}: {atype}' is undefined"
                )
            value = getattr(record, name)
            object.__setattr__(self, name, value)

    def _init_field_attributes(self, record: Dataset) -> None:
        # Initializes field attributes from the provided dataset. Copy
        # record field attributes to the instance, convert the data if
        # necessary.
        for name in get_annotations(self, DatasetView):
            value = getattr(self, name)
            if not isinstance(value, ViewField):
                continue
            value.set_entry(name)
            extracted_value = value(record)
            object.__setattr__(self, name, extracted_value)

    def _validate_preconditions(self) -> None:
        # Validates preconditions before initialization.
        if self.__dict__:
            raise TypeError(f"Datarecord {self.__class__.__name__} is corrupt")

    def _validate_postconditions(self) -> None:
        # Validates postconditions after initialization.
        all_annotations = get_annotations(self, DatasetView)
        for name, annotation in all_annotations.items():
            value = getattr(self, name)
            if validate_type(value, annotation):
                continue
            vtype = get_typehint(value)
            atype = get_annotated(annotation)
            raise TypeError(
                f"Attribute '{name}: {atype}' "
                f"does not match the given type '{vtype}'"
            )


_T = TypeVar("_T")


def netcdf_fragment(recordclass: type[_T]) -> type[_T]:
    class _FragmentRecord(DatasetView):
        pass

    _FragmentRecord.__annotations__ = recordclass.__annotations__.copy()
    _FragmentRecord.__module__ = recordclass.__module__
    _FragmentRecord.__name__ = recordclass.__name__
    _FragmentRecord.__qualname__ = recordclass.__qualname__

    for name in recordclass.__annotations__:
        if hasattr(recordclass, name):
            value = getattr(recordclass, name)
            setattr(_FragmentRecord, name, value)

    __post_init__ = getattr(recordclass, "__post_init__", None)
    if callable(__post_init__):
        setattr(_FragmentRecord, "__post_init__", __post_init__)

    original_module = getmodule(recordclass)
    setattr(original_module, recordclass.__name__, _FragmentRecord)

    return cast(type[_T], _FragmentRecord)



from collections.abc import Sequence
from pathlib import Path
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import (
    BoundaryNorm,
    Colormap,
    LinearSegmentedColormap,
    Normalize,
)
from numpy.typing import NDArray

from .colormap import EnhancementColormap
from .shared import DomainData, RGBValue
from .table import EnhacementTable
from .ticks import ColorbarTicks

ExtendMode = Literal["neither", "both", "min", "max"]


class EnhancementScale:

    cmap: Colormap
    cnorm: Normalize
    cticks: list[float]
    domain: DomainData
    extent: DomainData
    name: str
    ncolors: int
    offset: float

    def __init__(
        self,
        specs: EnhacementTable | EnhancementColormap,
        offset: float = 0.0,
        ncolors: int = 256,
        nticks: int = 16,
        mode: ExtendMode = "neither",
    ) -> None:
        if isinstance(specs, EnhacementTable):
            cmap, cnorm = self._create_from_table(specs, offset, ncolors, mode)
        else:
            cmap, cnorm = self._create_from_colormap(
                specs, offset, ncolors, mode
            )

        domain = self._actual_range(specs.domain, offset)
        extent = self._actual_range(specs.extent, offset)

        refs = ColorbarTicks(extent, nticks, step=0)

        self.cmap = cmap
        self.cnorm = cnorm
        self.cticks = refs.cticks
        self.domain = domain
        self.extent = extent
        self.name = specs.name
        self.ncolors = ncolors
        self.offset = offset

    def extend(self, mode: ExtendMode = "neither") -> None:
        self.cnorm = self._create_colorbar_norm(
            self.extent, self.ncolors, mode
        )

    @classmethod
    def from_colormap(
        cls,
        name: str,
        cmap_names: str | Sequence[str],
        keypoints: Sequence[float],
        offset: float = 0.0,
        ncolors: int = 256,
        nticks: int = 16,
        mode: ExtendMode = "neither",
    ) -> "EnhancementScale":
        cmap = EnhancementColormap(name, cmap_names, keypoints)

        return cls(cmap, offset, ncolors, nticks, mode)

    @classmethod
    def from_file(
        cls,
        palette_path: str | Path,
        stretching_path: str | Path = "",
        offset: float = 0.0,
        ncolors: int = 256,
        nticks: int = 16,
        mode: ExtendMode = "neither",
    ) -> "EnhancementScale":
        table = EnhacementTable.from_file(palette_path, stretching_path)

        return cls(table, offset, ncolors, nticks, mode)

    def set_ticks(self, nticks: int = 16, step: int = 0) -> None:
        refs = ColorbarTicks(self.extent, nticks, step)
        self.cticks = refs.cticks

    @staticmethod
    def _actual_range(extent: DomainData, offset: float) -> DomainData:
        return extent[0] + offset, extent[1] + offset

    @classmethod
    def _create_colorbar_norm(
        cls, extent: DomainData, ncolors: int, mode: ExtendMode
    ) -> BoundaryNorm:
        """
        Create a colorbar norm for the given temperature range.

        Creates BoundaryNorm with a list of levels evenly spaced between
        tmin and tmax.

        Parameters
        ----------
        extent: DomainData
            Tuple with minimum and maximum temperature.
        ncolors : int, optional
            Number of colormap levels.
        mode : ExtendMode
            Colorbar extend specification.

        Returns
        -------
        BoundaryNorm
            The colorbar norm.
        """
        tmin, tmax = extent
        vmin, vmax = round(tmin), round(tmax)

        nlevels = ncolors + 1
        levels = cls._linspace(vmin, vmax, nlevels)

        return BoundaryNorm(levels, ncolors=ncolors, clip=True, extend=mode)

    def _create_from_colormap(
        self,
        colormap: EnhancementColormap,
        offset: float,
        ncolors: int,
        mode: ExtendMode,
    ) -> tuple[LinearSegmentedColormap, BoundaryNorm]:
        extent = self._actual_range(colormap.extent, offset)

        cnorm = self._create_colorbar_norm(extent, ncolors, mode)

        tmin, tmax = extent
        total_range = tmax - tmin

        all_colors: list[NDArray[np.float64]] = []
        for i, name in enumerate(colormap.cmap_names):
            icmap = plt.get_cmap(name)
            start_keypoint = colormap.keypoints[i]
            end_keypoint = colormap.keypoints[i + 1]
            segment_range = end_keypoint - start_keypoint
            proportion = segment_range / total_range
            n_segment_colors = round(ncolors * proportion)
            colors = icmap(np.linspace(0, 1, n_segment_colors))
            all_colors.append(colors)

        colors = np.vstack(all_colors)

        cmap = LinearSegmentedColormap.from_list(
            colormap.name, colors, N=ncolors
        )

        icmap = plt.get_cmap(colormap.cmap_names[0])
        under_color = self._to_colortype(icmap.get_under())

        icmap = plt.get_cmap(colormap.cmap_names[-1])
        over_color = self._to_colortype(icmap.get_over())
        bad_color = self._to_colortype(icmap.get_bad())

        cmap.set_under(under_color)
        cmap.set_over(over_color)
        cmap.set_bad(bad_color)

        return cmap, cnorm

    def _create_from_table(
        self,
        table: EnhacementTable,
        offset: float,
        ncolors: int,
        mode: ExtendMode,
    ) -> tuple[LinearSegmentedColormap, BoundaryNorm]:
        extent = self._actual_range(table.extent, offset)

        cnorm = self._create_colorbar_norm(extent, ncolors, mode)

        cmap = LinearSegmentedColormap(
            table.name, table.color_table, N=ncolors
        )

        cmap.set_under(table.stock["background"])
        cmap.set_over(table.stock["foreground"])
        cmap.set_bad(table.stock["nan"])

        return cmap, cnorm

    @staticmethod
    def _linspace(start: float, stop: float, nlevels: int) -> list[float]:
        """
        Generates a list of evenly spaced numbers between start and stop.

        Parameters
        ----------
        start : float
            Initial value.
        stop : float
            Final value.
        nlevels : int
            Number of elements in the list.

        Returns
        -------
        list[float]
            A list with the generated numbers.
        """
        extent = stop - start
        nmax = nlevels - 1
        return [start + (i / nmax) * extent for i in range(nlevels)]

    @staticmethod
    def _to_colortype(color: NDArray[np.float64]) -> RGBValue:
        return color[0], color[1], color[2]



"""
Provide the Datasource abstract interface for handling datasources.

Classes:
    Datasource: Abstract a datasource object.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass(eq=False, frozen=True)
class Datasource(ABC):
    """
    Abstract a datasource object.

    This class defines the interface for a datasource object. The
    datasource is responsible for listing the contents of a directory in
    a remote location and for downloading files from that location.

    Methods
    -------
    download_file(file_path: str)
        Retrieve a file from the datasource and save it into the local
        repository.
    list_files(dir_path: str)
        List the contents of a remote directory.

    Attributes
    ----------
    base_url : str
        The base URL of the datasource. This is the URL where the
        datasource is located. The base URL is used to build the full
        URL to the files and directories.
    """

    base_url: str

    @abstractmethod
    def download_file(self, file_path: str) -> bytes:
        """
        Download a file from the datasource into the local repository.

        Get a file from a remote location or local repository. The path
        provided must be relative to the base URL and local repository
        root directory. The remote path is reconstructed in the local
        repository.

        Parameters
        ----------
        file_path : str
            The path to the remote file to be downloaded.

        Returns
        -------
        bytes
            The content of the file as a byte string.
        """

    @abstractmethod
    def list_files(self, dir_path: str) -> list[str]:
        """
        List the contents of a remote directory.

        List the contents of a directory in a remote location. The path
        is relative to the base URL.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to the base
            URL.

        Returns
        -------
        list[str]
            A list of file names in the directory.
        """



"""
Constants for GOES dataset file handling.

This module contains constants used for handling GOES dataset files,
including file suffixes, date formats, and date patterns.
"""

GOESR_FILE_SUFFIX: str = ".nc"

GOESR_FILE_DATE_FORMAT: str = "%Y%j%H%M%S%f"
GOESR_FILE_DATE_PATTERN: str = r"\d{14}"
GOESR_PATH_DATE_FORMAT: str = "%Y/%j/%H"



from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any

from netCDF4 import Dataset  # pylint: disable=no-name-in-module
from numpy import ndarray
from numpy.ma import MaskedArray

ConvertFn = Callable[..., Any] | type
FilterFn = Callable[..., Any]


class ClassField:
    """
    Placeholder for class attribute fields.

    Mark attributes that will be copied from the given value to the
    instance without any conversion.

    Attributes
    ----------
    value : Any
        The value to be copied.

    See Also
    --------
    computed :
        Placeholder for computed attributes.
    field :
        Placeholder for instance attributes.
    ViewField :
        Abstract base class for dataset view field placeholders.
    """

    value: Any

    def __init__(self, value: Any) -> None:
        """
        Initialize the class field.

        Parameters
        ----------
        value : Any
            The value to be copied.
        """
        self.value = value


def computed() -> Any:
    """
    Placeholder for computed instance attributes.

    Mark attributes that will be computed during the record
    initialization in the __post_init__ method.

    Returns
    -------
    Any
        An instance of 'ClassField(None)'.

    Examples
    --------
    >>> class Record(DatasetView):
    ...     field1: int = field(42)
    ...     field2: int = computed()
    ...

    The 'field1' attribute will be copied from the given value, while
    the 'field2' attribute will be computed during the record
    initialization.

    See Also
    --------
    attribute :
        Placeholder for dataset view attribute fields.
    data :
        Placeholder for dataset view array data fields.
    dimension :
        Placeholder proxy for dataset view dimension fields.
    field :
        Placeholder for instance attributes.
    variable :
        Placeholder proxy for dataset view variable fields.
    """
    return ClassField(None)


def field(value: Any) -> Any:
    """
    Placeholder for instance attributes.

    Mark attributes that will be copied from the given value to the
    instance without any conversion.

    Parameters
    ----------
    value : Any
        The value to be copied.

    Returns
    -------
    Any
        An instance of 'ClassField(value)'.

    Examples
    --------
    >>> class Record(DatasetView):
    ...     field1: int = field(42)
    ...     field2: int = computed()
    ...

    The 'field1' attribute will be copied from the given value, while
    the 'field2' attribute will be computed during the record
    initialization.

    See Also
    --------
    attribute :
        Placeholder for dataset view attribute fields.
    computed :
        Placeholder for computed attributes.
    data :
        Placeholder for dataset view array data fields.
    dimension :
        Placeholder proxy for dataset view dimension fields.
    variable :
        Placeholder proxy for dataset view variable fields.
    """
    return ClassField(value)


class ViewField(ABC):
    """
    Abstract base class for dataset view field placeholders.

    Attributes
    ----------
    id : str | None
        The name of the field to be copied.
    entry : str | None
        The name of the field entry to be copied.
    convert : ConvertFn
        The conversion function to be applied to the field entry value.

    See Also
    --------
    AttributeField :
        Represent a placeholder for dataset view attribute fields.
    DataField :
        Represent a placeholder for dataset view array data fields.
    DimensionField :
        Represent a placeholder for dataset view dimension fields.
    VariableField :
        Represent a placeholder for dataset view variable fields.
    """

    id: str | None
    entry: str | None
    convert: ConvertFn

    def __init__(
        self,
        id: str | None,
        entry: str | None,
        convert: ConvertFn | None,
    ) -> None:
        """
        Initialize the view field.

        Parameters
        ----------
        id : str | None
            The name of the field to be copied.
        entry : str | None
            The name of the field entry to be copied.
        convert : ConvertFn | None
            The conversion function to be applied to the field entry
            value, default to the identity function.
        """
        self.id = id
        self.entry = entry

        def _identity(x: Any) -> Any:
            return x

        self.convert = convert or _identity

    @abstractmethod
    def __call__(self, dataset: Dataset) -> Any:
        """
        Copy the field entry value from the dataset.

        This method must be implemented by the subclasses to copy the
        field entry value from the dataset.

        Parameters
        ----------
        dataset : Dataset
            The dataset to copy the field entry value from.

        Returns
        -------
        Any
            The field entry value copied from the dataset.
        """

    def set_entry(self, entry: str) -> None:
        """
        Set the field entry name.

        This method is used to set the field entry name when it is not
        provided explicitly. It does not override the existing entry
        name.

        Parameters
        ----------
        entry : str
            The name of the field entry to be copied.
        """
        self.entry = self.entry or entry


class AttributeField(ViewField):
    """
    Represent a placeholder for dataset view attribute field.

    Mark attributes that will be copied from existing attributes in the
    dataset to the instance with some optional conversion.

    See Also
    --------
    attribute :
        Placeholder for dataset view attribute fields.
    DataField :
        Represent a placeholder for dataset view array data fields.
    DimensionField :
        Represent a placeholder for dataset view dimension fields.
    VariableField :
        Represent a placeholder for dataset view variable fields.
    ViewField :
        Abstract base class for dataset view field placeholders.
    """

    def __init__(self, entry: str | None, convert: ConvertFn | None) -> None:
        """
        Initialize the attribute field.

        Parameters
        ----------
        entry : str | None
            The name of the attribute to be copied.
        convert : ConvertFn | None
            The conversion function to be applied to the attribute
            value, default to the identity function.

        Returns
        -------
        Any
            An instance of 'AttributeField(entry, convert)'.

        Raises
        ------
        ValueError
            If the attribute name is not provided or unknown.
        """
        super().__init__(None, entry, convert)

    def __call__(self, dataset: Dataset) -> Any:
        if self.entry is None:
            raise ValueError("Attribute name is required")
        try:
            value_ = getattr(dataset, self.entry)
            return self.convert(value_)
        except AttributeError as error:
            raise ValueError(f"Unknown attribute '{self.entry}'") from error


def attribute(
    entry: str | None = None, *, convert: ConvertFn | None = None
) -> Any:
    """
    Placeholder for dataset view attribute field.

    Mark attributes that will be copied from existing attributes in the
    dataset to the instance with some optional conversion.

    Parameters
    ----------
    entry : str | None, optional
        The name of the attribute to be copied.
    convert : ConvertFn | None, optional
        The conversion function to be applied to the attribute value,
        default to the identity function.

    Returns
    -------
    Any
        An instance of 'AttributeField(entry, convert)'.

    Examples
    --------
    >>> class Record(DatasetView):
    ...     attr0: str = attribute()  # use ds.attr0
    ...     attribute1: int = attribute("attr1")  # use ds.attr0
    ...     attribute2: str = attribute("attr2", convert=str)  # use str(ds.attr2)
    ...

    See Also
    --------
    AttributeField :
        Represent a placeholder for dataset view attribute fields.
    computed :
        Placeholder for computed attributes.
    data :
        Placeholder for dataset view array data fields.
    dimension :
        Placeholder proxy for dataset view dimension fields.
    field :
        Placeholder for instance attributes.
    variable :
        Placeholder proxy for dataset view variable fields.
    """
    return AttributeField(entry, convert)


class DimensionField(ViewField):
    """
    Represent a placeholder for dataset view dimension fields.

    Mark attributes that will be copied from existing dimensions in the
    dataset to the instance with some optional conversion.

    See Also
    --------
    dimension :
        Placeholder proxy for dataset view dimension fields.
    AttributeField :
        Represent a placeholder for dataset view attribute fields.
    DataField :
        Represent a placeholder for dataset view array data fields.
    DimensionProxy :
        Proxy for dataset view dimension fields.
    VariableField :
        Represent a placeholder for dataset view variable fields.
    ViewField :
        Abstract base class for dataset view field placeholders.
    """

    def __init__(
        self, id: str | None, entry: str | None, convert: ConvertFn | None
    ) -> None:
        """
        Initialize the dimension field.

        Parameters
        ----------
        id : str | None
            The name of the dimension to be copied.
        entry : str | None
            The name of the dimension entry to be copied.
        convert : ConvertFn | None
            The conversion function to be applied to the dimension entry
            value, default to the identity function.
        """
        super().__init__(id, entry, convert)

    def __call__(self, dataset: Dataset) -> Any:
        if self.id is None:
            raise ValueError("Dimension id is required")
        if self.entry is None:
            raise ValueError("Dimension entry is required")
        return self._get_dimension(dataset, self.id, self.entry)

    def _get_dimension(self, dataset: Dataset, id: str, entry: str) -> Any:
        try:
            dimension_ = dataset.dimensions[id]
            value_ = getattr(dimension_, entry)
            return self.convert(value_)
        except IndexError as error:
            raise ValueError(f"Unknown dimension '{id}'") from error
        except AttributeError as error:
            raise ValueError(
                f"Unknown dimension entry '{id}:{entry}'"
            ) from error

    def set_entry(self, entry: str) -> None:
        """
        Set the dimension field name.

        This method is used to set the field entry name when it is not
        provided explicitly. It does not override the existing entry
        name.

        Parameters
        ----------
        entry : str
            The name of the dimension field to be copied.
        """
        self.id = self.id or entry


class DimensionProxy:
    """
    Proxy class for dataset view dimension fields.

    Attributes
    ----------
    id : str | None
        The name of the dimension to be copied.

    See Also
    --------
    dimension :
        Placeholder proxy for dataset view dimension fields.
    DimensionField :
        Represent a placeholder for dataset view dimension fields.
    """

    id: str | None = None

    def __init__(self, id: str | None) -> None:
        """
        Initialize the dimension proxy.

        Parameters
        ----------
        id : str | None
            The name of the dimension to be copied.
        """
        self.id = id

    def __call__(self, id: str) -> "DimensionProxy":
        return DimensionProxy(id)

    def name(self, *, convert: ConvertFn | None = None) -> Any:
        """
        Placeholder for dataset dimension name attribute.

        Parameters
        ----------
        convert : ConvertFn | None, optional
            The conversion function to be applied to the dimension name,
            default to the identity function.

        Returns
        -------
        Any
            An instance of 'DimensionField(id, "name", convert)'.
        """
        return DimensionField(self.id, "name", convert)

    def size(self, *, convert: ConvertFn | None = None) -> Any:
        """
        Placeholder for dataset dimension size attribute.

        Parameters
        ----------
        convert : ConvertFn | None, optional
            The conversion function to be applied to the dimension size,
            default to the identity function.

        Returns
        -------
        Any
            An instance of 'DimensionField(id, "size", convert)'.
        """
        return DimensionField(self.id, "size", convert)


dimension = DimensionProxy(None)
"""
Placeholder proxy for dataset view dimension fields.

Examples
--------
>>> class Record(DatasetView):
...     row: int = dimension.size()  # use ds.dimensions["row"].size
...     time_value: float = dimension("time").size(float)
...     time_name: str = dimension("time").name()
...

See Also
--------
attribute :
    Placeholder for dataset view attribute fields.
computed :
    Placeholder for computed attributes.
data :
    Placeholder for dataset view array data fields.
field :
    Placeholder for instance attributes.
variable :
    Placeholder proxy for dataset view variable fields.
DimensionField :
    Represent a placeholder for dataset view dimension fields.
DimensionProxy :
    Proxy for dataset view dimension fields.
"""


ARRAY_PREFIX = "array:"
ARRAY = "array:*"
DATA = "array:data"
NODATA = "array:fill_value"
MASK = "array:mask"


class VariableField(ViewField):
    """
    Represent a placeholder for dataset view variable fields.

    Mark attributes that will be copied from existing variables in the
    dataset to the instance with some optional filtering and conversion.

    Attributes
    ----------
    filter : FilterFn
        The filter function to be applied to the variable entry.

    See Also
    --------
    variable :
        Placeholder proxy for dataset view variable fields.
    VariableProxy :
        Proxy for dataset view variable fields.
    AttributeField :
        Represent a placeholder for dataset view attribute fields.
    DataField :
        Represent a placeholder for dataset view array data fields.
    DimensionField :
        Represent a placeholder for dataset view dimension fields.
    ViewField :
        Abstract base class for dataset view field placeholders.
    """

    filter: FilterFn

    def __init__(
        self,
        id: str | None,
        entry: str | None,
        filter: FilterFn | None,
        convert: ConvertFn | None,
    ) -> None:
        """
        Initialize the variable field.

        Parameters
        ----------
        id : str | None
            The name of the variable to be copied.
        entry : str | None
            The name of the variable entry to be copied.
        filter : FilterFn | None
            The filter function to be applied to the variable entry.
        convert : ConvertFn | None
            The conversion function to be applied to the
            variable entry value.
        """
        super().__init__(id, entry, convert)

        def _extract_all(x: Any) -> Any:
            return x[:]

        self.filter = filter or _extract_all

    def __call__(self, dataset: Dataset) -> Any:
        if self.id is None:
            raise ValueError("Variable id is required")
        if self.entry is None:
            raise ValueError("Variable entry is required")
        return self._get_variable(dataset, self.id, self.entry)

    def _extract_array(self, variable_: Any, entry: str) -> Any:
        array_value = self.filter(variable_)
        if not isinstance(array_value, ndarray):
            raise TypeError("Unsuported variable entry type")
        if isinstance(array_value, MaskedArray):
            if entry == DATA:
                return array_value.data
            if entry == MASK:
                return array_value.mask
            if entry == NODATA:
                return array_value.fill_value
        if entry == ARRAY:
            return array_value
        raise AttributeError("Unknown array entry")

    def _get_variable(self, dataset: Dataset, id: str, entry: str) -> Any:
        try:
            variable_ = dataset.variables[id]
            if entry.startswith(ARRAY_PREFIX):
                value_ = self._extract_array(variable_, entry)
            else:
                value_ = getattr(variable_, entry)
            return self.convert(value_)
        except IndexError as error:
            raise ValueError(f"Unknown variable '{id}'") from error
        except AttributeError as error:
            raise ValueError(
                f"Unknown variable entry '{id}:{entry}'"
            ) from error
        except TypeError as error:
            raise ValueError(
                f"Unexpected variable entry type '{id}:{entry}'"
            ) from error


class VariableProxy:
    """
    Proxy class for dataset view variable fields.

    Attributes
    ----------
    id : str
        The name of the variable to be copied.

    See Also
    --------
    variable :
        Placeholder proxy for dataset view variable fields.
    VariableField :
        Represent a placeholder for dataset view variable fields.
    """

    id: str

    def __init__(self, id: str) -> None:
        """
        Initialize the variable proxy.

        Parameters
        ----------
        id : str
            The name of the variable to be copied.
        """
        self.id = id

    def array(
        self,
        entry: str | None = None,
        *,
        filter: FilterFn | None = None,
        convert: ConvertFn | None = None,
    ) -> Any:
        """
        Placeholder for dataset view variable array field.

        Parameters
        ----------
        entry : str | None, optional
            The name of the variable entry to be copied, choose: 'data',
            'mask', 'fill_value', or '*' (default).
        filter : FilterFn | None, optional
            The filter function to be applied to the variable entry.
        convert : ConvertFn | None, optional
            The conversion function to be applied to the variable entry
            value.

        Returns
        -------
        Any
            An instance of 'VariableField(id, entry, filter, convert)'.

        Raises
        ------
        ValueError
            If the entry is not valid or unknown.
        """
        if entry not in {"*", "data", "mask", "fill_value", None}:
            raise ValueError(
                "Invalid 'entry', choose: '*', 'data', 'mask', or 'fill_value'"
            )
        entry = f"array:{entry or '*'}"
        return VariableField(self.id, entry, filter, convert)

    def attribute(
        self, entry: str | None = None, *, convert: ConvertFn | None = None
    ) -> Any:
        """
        Placeholder for dataset view variable attribute field.

        Parameters
        ----------
        entry : str | None, optional
            The name of the attribute to be copied.
        convert : ConvertFn | None, optional
            The conversion function to be applied to the
            attribute value.

        Returns
        -------
        Any
            An instance of 'VariableField(id, entry, None, convert)'.
        """
        return VariableField(self.id, entry, None, convert)

    def data(
        self,
        *,
        filter: FilterFn | None = None,
        convert: ConvertFn | None = None,
    ) -> Any:
        """
        Placeholder for the data component of array variable field.

        Parameters
        ----------
        filter : FilterFn | None, optional
            The filter function to be applied to the variable entry.
        convert : ConvertFn | None, optional
            The conversion function to be applied to the variable entry
            value.

        Returns
        -------
        Any
            An instance of 'VariableField(id, "data", filter, convert)'.
        """
        return self.array("data", filter=filter, convert=convert)

    def fill_value(
        self,
        *,
        convert: ConvertFn | None = None,
    ) -> Any:
        """
        Placeholder for the fil value component of array variable field.

        Parameters
        ----------
        convert : ConvertFn | None, optional
            The conversion function to be applied to the variable entry
            value.

        Returns
        -------
        Any
            An instance of 'VariableField(id, "fill_value", filter,
            convert)'.
        """
        return self.array("fill_value", filter=None, convert=convert)

    def mask(
        self,
        *,
        filter: FilterFn | None = None,
        convert: ConvertFn | None = None,
    ) -> Any:
        """
        Placeholder for the mask component of array variable field.

        Parameters
        ----------
        filter : FilterFn | None, optional
            The filter function to be applied to the variable entry.
        convert : ConvertFn | None, optional
            The conversion function to be applied to the variable entry
            value.

        Returns
        -------
        Any
            An instance of 'VariableField(id, "mask", filter, convert)'.
        """
        return self.array("mask", filter=filter, convert=convert)


def variable(name: str) -> VariableProxy:
    """
    Placeholder proxy for dataset view variable fields.

    Parameters
    ----------
    name : str
        The name of the variable to be copied.

    Returns
    -------
    VariableProxy
        An instance of 'VariableProxy(name)'.

    Examples
    --------
    >>> image = variable("image")
    ...
    ... class Record(DatasetView):
    ...     image: NDArray[float32] = image.array()  # or image.array("data")
    ...     # retrieve ds.variables["image"][:].data
    ...
    ...     mask: NDArray[bool] = image.array()  # or image.array("mask")
    ...     # retrieve ds.variables["image"][:].data
    ...
    ...     fill_value: float32 = image.array("fill_value")
    ...     # retrieve ds.variables["image"][0:1,0:1].fill_value
    ...
    ...     thumb: NDArray[uint8] = image.array(
    ...         filter=lambda x: x[::4, ::4],
    ...         convert=lambda x: x.astype(uint8),
    ...     )
    ...     # retrieve ds.variables["image"][::4, ::4].data.astype(uint8)
    ...
    ...     long_name: str = image.attribute()  # or image.attribute("long_name")
    ...     # retrieve ds.variables["image"].long_name
    ...
    ...     rows: int = image.scalar()  # or image.scalar("rows")
    ...     # retrieve ds.variables["rows"][0].data
    ...

    See Also
    --------
    attribute :
        Placeholder for dataset view attribute fields.
    computed :
        Placeholder for computed attributes.
    data :
        Placeholder for dataset view array data fields.
    dimension :
        Placeholder proxy for dataset view dimension fields.
    field :
        Placeholder for instance attributes.
    VariableField :
        Represent a placeholder for dataset view variable fields.
    VariableProxy :
        Proxy for dataset view variable fields.
    """
    return VariableProxy(name)


class DataField(VariableField):
    """
    Represent a placeholder for dataset view array data fields.

    Mark attributes that will be copied from existing variables in the
    dataset to the instance with some optional filtering and conversion.

    See Also
    --------
    variable :
        Placeholder proxy for dataset view variable fields.
    VariableProxy :
        Proxy for dataset view variable fields.
    AttributeField :
        Represent a placeholder for dataset view attribute fields.
    DimensionField :
        Represent a placeholder for dataset view dimension fields.
    VariableField :
        Represent a placeholder for dataset view variable fields.
    ViewField :
        Abstract base class for dataset view field placeholders.
    """

    def __init__(
        self,
        id: str | None,
        filter: FilterFn | None,
        convert: ConvertFn | None,
    ) -> None:
        """
        Initialize the variable data field.

        Parameters
        ----------
        id : str | None
            The name of the variable to be copied.
        filter : FilterFn | None
            The filter function to be applied to the variable entry.
        convert : ConvertFn | None
            The conversion function to be applied to the
            variable entry value.
        """
        super().__init__(id, DATA, filter, convert)

    def set_entry(self, entry: str) -> None:
        """
        Set the variable field name.

        This method is used to set the field entry name when it is not
        provided explicitly. It does not override the existing entry
        name.

        Parameters
        ----------
        entry : str
            The name of the variable field to be copied.
        """
        self.id = self.id or entry


def data(
    id: str | None = None,
    *,
    filter: FilterFn | None = None,
    convert: ConvertFn | None = None,
) -> Any:
    """
    Placeholder for dataset view array data fields.

    Parameters
    ----------
    id : str | None, optional
        The name of the variable.
    filter : FilterFn | None, optional
        The filter function to be applied to the variable entry.
    convert : ConvertFn | None, optional
        The conversion function to be applied to the variable entry
        value.

    Returns
    -------
    Any
        An instance of 'DataField(id, filter, convert)'.

    See Also
    --------
    attribute :
        Placeholder for dataset view attribute fields.
    computed :
        Placeholder for computed attributes.
    dimension :
        Placeholder proxy for dataset view dimension fields.
    field :
        Placeholder for instance attributes.
    variable :
        Placeholder proxy for dataset view variable fields.
    DataField :
        Represent a placeholder for dataset view array data fields.
    """
    return DataField(id, filter, convert)


def scalar(id: str | None = None, *, convert: ConvertFn | None = None) -> Any:
    """
    Placeholder for dataset view scalar variable field.

    This effectively retrieves the first element of the variable
    array.

    Parameters
    ----------
    id : str | None, optional
        The name of the variable.
    convert : ConvertFn | None, optional
        The conversion function to be applied to the variable entry
        value.

    Returns
    -------
    Any
        The instance of 'VariableField' returned by 'data(id,
        filter=*, convert=convert)'.
    """

    def filter_(x: Any) -> Any:
        return x[:1]

    def convert_(x: Any) -> Any:
        return convert(x[0]) if convert else x[0]

    return data(id, filter=filter_, convert=convert_)



"""
Calculate latitude and longitude grids data using pyproj.

Notes
-----
See GOES-R Product User Guide (PUG) Volume 5 (L2 products) Section 4.2.8
for details & example of calculations.

Functions
---------
calculate_latlon_grid_pyproj
    Calculate latitude and longitude grids using the pyproj package.
"""

from numpy import float32, meshgrid

from ..utils.array import ArrayFloat32, ArrayFloat64
from .helpers import make_consistent
from .geostationary_parameters import GeostationaryParameters


def geos_to_latlon_grid_pyproj(
    projection_info: GeostationaryParameters,
) -> tuple[ArrayFloat32, ArrayFloat32]:
    """
    Calculate latitude and longitude grids using the pyproj package.

    Calculate latitude and longitude from GOES ABI fixed grid projection
    data. GOES ABI fixed grid projection is a map projection relative to
    the GOES satellite.

    Units: latitude in °N (°S < 0), longitude in °E (°W < 0)

    Parameters
    ----------
    projection_info : GOESProjection
        Object containing the satellite's projection information and
        GOES ABI fixed grid data.

    Returns
    -------
    tuple[ArrayFloat32, ArrayFloat32]
        A tuple containing the latitude and longitude grid data.
    """
    try:
        from pyproj import Proj  # , Transformer
    except ImportError as error:
        raise ImportError(
            "The 'pyproj' package is required for this functionality."
        ) from error

    geos_proj = Proj(
        proj="geos",
        h=projection_info.orbit.perspective_point_height,
        lon_0=projection_info.orbit.longitude_of_projection_origin,
        sweep=projection_info.orbit.sweep_angle_axis,
        a=projection_info.globe.semi_major_axis,
        b=projection_info.globe.semi_minor_axis,
        rf=projection_info.globe.inverse_flattening,
    )

    x_m: ArrayFloat64
    y_m: ArrayFloat64
    x_m, y_m = meshgrid(projection_info.x_m, projection_info.y_m)

    abi_lon: ArrayFloat64
    abi_lat: ArrayFloat64
    abi_lon, abi_lat = geos_proj(x_m, y_m, inverse=True)

    # The above is equivalent to the following:
    #
    # plate_carree_proj = Proj(
    #     proj="latlong",
    #     a=projection_info.globe.semi_major_axis,
    #     b=projection_info.globe.semi_minor_axis,
    #     rf=projection_info.globe.inverse_flattening,
    # )
    # transformer = Transformer.from_proj(geos_proj, plate_carree_proj)
    #
    # abi_lon, abi_lat = transformer.transform(x_m, y_m)

    abi_lon, abi_lat = make_consistent(abi_lon, abi_lat)

    return abi_lat.astype(float32), abi_lon.astype(float32)



"""
GOES Image data extraction.

This module provides classes to extract and represent image information
from GOES satellite netCDF data files.

Classes
-------
GOESImage
    Represent a GOES satellite image data.
GOESImageMetadata
    Represent GOES image metadata attributes.
"""

from numpy import bool_, float32, int8, int16, int32, uint16
from numpy.typing import NDArray

from ..netcdf import DatasetView, scalar, variable

cmip = variable("CMI")


class GOESImage(DatasetView):
    """
    Represent a GOES satellite image data.

    Hold data for the Cloud and Moisture Imagery (CMI) bands.

    Attributes
    ----------
    band_id : int32
        The ID of the band.
    band_wavelength : float32
        The wavelength of the band.
    data : NDArray[float32]
        The image data.
    mask : NDArray[bool]
        The array containing the mask values indicating invalid data
        points.
    fill_value : uint16
        The fill value used for missing or invalid data points.
    """

    band_id: int32 = scalar()
    band_wavelength: float32 = scalar()

    data: NDArray[float32] = cmip.array()
    mask: NDArray[bool_] = cmip.array(entry="mask")
    fill_value: uint16 = cmip.array(entry="fill_value")


class GOESImageMetadata(DatasetView):
    """
    Represent GOES image metadata attributes.

    Attributes
    ----------
    long_name : str
        A descriptive name for the dataset.
    standard_name : str
        A standard name for the dataset.
    sensor_band_bit_depth : int8
        The bit depth of the sensor band.
    valid_range : NDArray[int16]
        The valid range of data values.
    scale_factor : float32
        The scale factor to apply to the data values.
    add_offset : float32
        The offset to add to the data values.
    units : str
        The units of the data values.
    resolution : str
        The spatial resolution of the data.
    grid_mapping : str
        The grid mapping information.
    """

    long_name: str = cmip.attribute()
    standard_name: str = cmip.attribute()
    sensor_band_bit_depth: int8 = cmip.attribute()
    valid_range: NDArray[int16] = cmip.attribute()
    scale_factor: float32 = cmip.attribute()
    add_offset: float32 = cmip.attribute()
    units: str = cmip.attribute()
    resolution: str = cmip.attribute()
    grid_mapping: str = cmip.attribute()



"""
Provide locator for GOES-R Series imagery dataset's ABI products.

Classes:
    - GOESProductLocatorABIPP: All primary ABI products.
    - GOESProductLocatorCMIP: ABI Cloud and Moisture Imagery Product
      (CMIP).
    - GOESProductLocatorRad: ABI Radiance Product (Rad).
"""

from .locator_abi import GOESProductLocatorABI


class GOESProductLocatorABIPP(GOESProductLocatorABI):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: All primary ABI products.
    """

    AVAILABLE_CHANNELS: set[str] = {f"C{idn:02d}" for idn in range(1, 17)}

    AVAILABLE_PRODUCTS: dict[str, str] = {
        "CMIP": "Cloud and Moisture Imagery Product",
        "Rad": "Radiance",
    }

    PRODUCT_RAD: str = "Rad"

    def __init__(
        self, name: str, scene: str, channels: str | list[str], origin: str
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI primary product
        locator object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        channels : str | list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C13".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.

        Raises
        ------
        ValueError
            If the provided product name is invalid.
        """
        if isinstance(channels, str):
            channels = [channels]

        if not channels:
            raise ValueError(
                f"Primary ABI product '{name}' "
                "does require channel specification"
            )

        self._validate_channels(channels, self.AVAILABLE_CHANNELS)

        self._validate_product(name, self.AVAILABLE_PRODUCTS)

        level: str = (
            self.LEVEL_RAD if name == self.PRODUCT_RAD else self.LEVEL_NOT_RAD
        )

        super().__init__(
            name=name,
            level=level,
            scene=scene,
            channels=channels,
            origin=origin,
        )


class GOESProductLocatorRad(GOESProductLocatorABIPP):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Radiance (Rad).
    """

    PRODUCT_NAME_RAD: str = GOESProductLocatorABIPP.PRODUCT_RAD

    def __init__(
        self, scene: str, channels: str | list[str], origin: str
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI primary product
        locator object.

        Parameters
        ----------
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        channels : str | list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C13".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        super().__init__(
            name=self.PRODUCT_NAME_RAD,
            scene=scene,
            channels=channels,
            origin=origin,
        )


class GOESProductLocatorCMIP(GOESProductLocatorABIPP):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Cloud and Moisture Imagery Product (CMIP).
    """

    PRODUCT_NAME_CMIP: str = "CMIP"

    def __init__(
        self, scene: str, channels: str | list[str], origin: str
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI primary product
        locator object.

        Parameters
        ----------
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        channels : str | list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C13".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        super().__init__(
            name=self.PRODUCT_NAME_CMIP,
            scene=scene,
            channels=channels,
            origin=origin,
        )



"""
Provide product locator for satellite imagery dataset consumers.

Classes:
    - ProductLocator: Abstract a product locator for satellite imagery
      dataset consumers.
"""

from abc import ABC, abstractmethod
from datetime import datetime


class ProductLocator(ABC):
    """
    Abstract a product locator for satellite imagery dataset consumers.

    This abstract base class defines the product locator interface for
    consumers of satellite imagery datasets. Subclasses instances are
    responsible for:

    1) Generating a list of folder paths based on the dataset's
       directory structure and naming conventions, product details,
       and a specified date range. The generated paths must cover
       the time interval defined by the dataset's directory temporal
       granularity within the required period; paths to the folders
       containing the initial and final dates must be included in
       the list.

    2) Verifying if a given filename matches the product filename
       pattern based on the dataset file's naming conventions and
       product specifications.

    3) Extracting the corresponding UTC `datetime` information from a
       valid product's filename.

    Notes
    -----
    Subclasses must implement the following methods: `get_base_url()`,
    `get_datetime()`, `get_paths()`, and `match()`; refer to their
    individual documentation for details.

    Important
    ---------
    It is assumed that the timestamp extracted from the product's
    filename is always in UTC timezone. Consumers of the product
    utilities should be aware of this assumption. Users can convert the
    `datetime` object to the desired timezone if needed. Implementors
    of the `ProductLocator` interface should ensure that the extracted
    `datetime` object is in UTC timezone in the case that the filename's
    timestamp is in a different timezone.

    Methods
    -------
    get_base_url(datasource: str) -> tuple[str, ...]:
        Get the base URL for the dataset's products.
    get_datetime(filename: str) -> datetime:
        Extracts the `datetime` from the product's filename.
    get_paths(datetime_ini: datetime, datetime_fin: datetime) -> list[str]:
        Generate a list of paths containing the product files for the
        specified date range.
    match(filename: str) -> bool:
        Verify if a given filename matches the product's filename
        pattern.
    """

    @abstractmethod
    def get_base_url(self, datasource: str) -> tuple[str, ...]:
        """
        Get the base URL for the dataset's products.

        This method returns the base URL for the dataset's products. The
        base URL is used to construct the full URL to the dataset's
        product files.

        Parameters
        ----------
        datasource : str
            The datasource identifier. This parameter is used to
            determine the base URL for the dataset's products. E.g.
            'AWS', 'GCP', 'NOAA'.

        Returns
        -------
        tuple[str, ...]
            A tuple where the first element is the base URL for the
            dataset's products. Additional elements may be included if
            the dataset is distributed across multiple sources or the
            datasource requires additional information.
        """

    @abstractmethod
    def get_datetime(self, filename: str) -> datetime:
        """
        Extract the `datetime` from the product's filename.

        This method parses the given filename and convert it to the
        corresponding `datetime` object from the product's filename
        using the dataset's date and time format conventions.

        Parameters
        ----------
        filename : str
            The filename from which to extract the `datetime`.

        Returns
        -------
        datetime
            The `datetime` extracted from the filename.

        Notes
        -----
        ValueError is raised if the filename does not match the expected
        pattern or if the dataset's datetime format specification is
        ill-formed.
        """

    @abstractmethod
    def get_paths(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> list[str]:
        """
        Generate a list of dataset directory paths.

        This method generates a list of directory paths within the
        dataset based on the folder structure and naming conventions,
        temporal granularity, and the specified date range. Paths to the
        folders containing the initial and final dates are included in
        the list.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime for the desired data.
        datetime_fin : datetime
            The final datetime for the desired data.

        Returns
        -------
        list[str]
            A list of strings representing the paths to dataset
            directories containing the product files for the
            specified date range.
        """

    @abstractmethod
    def match(self, filename: str) -> bool:
        """
        Verify if a filename matches the product's filename pattern.

        This method checks the given filename against the dataset
        product filename pattern based on the dataset's naming
        conventions and product specifications.

        Parameters
        ----------
        filename : str
            The filename to match against the pattern.

        Returns
        -------
        bool
            True if the filename matches the pattern, False otherwise.
        """



"""
Export the classes for the datasource objects.

The datasource objects are responsible for listing the contents of a
directory in a remote location and for downloading files from that
location.

Classes
-------
Datasource
    Abstract a datasource object.
DatasourceAWS
    A datasource object for an AWS S3 bucket.
DatasourceCache
    A datasource object that caches the files.
DatasourceNCEI
    A datasource object for an HTTP server.
DatasourceLocal
    A datasource object for a local directory.
"""

from .datasource import Datasource
from .datasource_aws import DatasourceAWS
from .datasource_cache import DatasourceCache
from .datasource_local import DatasourceLocal
from .datasource_ncei import DatasourceNCEI

__all__ = [
    "Datasource",
    "DatasourceAWS",
    "DatasourceCache",
    "DatasourceNCEI",
    "DatasourceLocal",
]



import math
from re import search
from typing import Any

from netCDF4 import Dataset  # pylint: disable=no-name-in-module

from ..netcdf import DatasetView, HasStrHelp, attribute
from .constants import NA
from .databook_gc import (
    PIXELS_PER_DEGREE,
    abstract_gridsat_gc,
    channel_correspondence,
    channel_description_gc,
    dataset_name_gridsat_gc,
    geospatial_resolution_deg,
    geospatial_resolution_km,
    measurement_range_lower_bound,
    measurement_range_upper_bound,
    measurement_units,
    platform_gridsat_gc,
    square_igfov_at_nadir,
    wavelength_range_lower_bound,
    wavelength_range_upper_bound,
)

NAN_TUPLE = math.nan, math.nan


class DatabookMetadata(HasStrHelp):

    dataset: str = NA
    abstract: str = NA
    description: str = NA
    channel_id: str = NA
    channel: str = NA
    geospatial_resolution_deg: tuple[float, float] = NAN_TUPLE
    geospatial_resolution_km: tuple[float, float] = NAN_TUPLE
    measurement_bounds: tuple[float, float] = NAN_TUPLE
    measurement_units: str = NA
    pixels_per_degree: int = PIXELS_PER_DEGREE
    square_fov_at_nadir: float = math.nan
    wavelength: float = math.nan
    wavelength_bounds: tuple[float, float] = NAN_TUPLE

    def __init__(self, channel: str, platform: str) -> None:
        origin = platform_gridsat_gc[platform]
        channel_orig = channel_correspondence[origin][channel]

        self.dataset = dataset_name_gridsat_gc
        self.abstract = abstract_gridsat_gc

        if channel_orig == 0:
            self.channel += (
                f"GridSat '{channel}' is not supported by {platform} origin"
            )
            return

        self.description = channel_description_gc[channel]

        self.channel_id = channel
        self.channel = f"Channel {channel_orig}"

        self.geospatial_resolution_deg = geospatial_resolution_deg
        self.geospatial_resolution_km = geospatial_resolution_km

        measurement_lo = measurement_range_lower_bound[origin][channel_orig]
        measurement_up = measurement_range_upper_bound[origin][channel_orig]
        self.measurement_bounds = measurement_lo, measurement_up

        self.measurement_units = measurement_units[channel_orig]

        wavelength_lo = wavelength_range_lower_bound[origin][channel_orig]
        wavelength_up = wavelength_range_upper_bound[origin][channel_orig]
        self.wavelength_bounds = wavelength_lo, wavelength_up

        self.wavelength = 0.5 * (wavelength_lo + wavelength_up)

        self.square_fov_at_nadir = square_igfov_at_nadir[origin][channel_orig]


class DatasetMetadata(DatasetView):

    title: str = attribute()
    id: str = attribute()
    summary: str = attribute()
    conventions: str = attribute("Conventions")
    license: str = attribute()
    processing_level: str = attribute()
    product_version: str = attribute()
    project: str = attribute()
    institution: str = attribute()
    comment: str = attribute()
    platform: str = attribute()
    instrument: str = attribute()
    keywords: str = attribute()
    platform_vocabulary: str = attribute()
    sensor_vocabulary: str = attribute()
    keywords_vocabulary: str = attribute()
    naming_authority: str = attribute()
    standard_name_vocabulary: str = attribute()
    metadata_link: str = attribute()
    ncei_template_version: str = attribute()
    date_created: str = attribute()
    date_modified: str = attribute()
    projection: str = attribute("Projection")
    time_coverage_start: str = attribute()
    time_coverage_end: str = attribute()
    history: str = attribute()

    @property
    def origin(self) -> str:
        return (
            match[0]
            if (match := search(r"GOES-\d{1,2}", self.platform))
            else ""
        )


class GSDatasetMetadata(DatabookMetadata, DatasetMetadata):

    def __init__(self, record: Dataset, channel: str) -> None:
        DatasetMetadata.__init__(self, record, channel=channel)

    def __post_init__(self, record: Dataset, **kwargs: Any) -> None:
        channel: str = kwargs["channel"]
        DatabookMetadata.__init__(self, channel, self.origin)



"""
Provide locator for GOES-R Series imagery dataset's ABI products.

Classes:
    - GOESProductLocatorABIPM: Multi-band primary ABI products.
    - GOESProductLocatorMCMIP: Multi-band CMIP Product.
"""

from .locator_nso import GOESProductLocatorABINSO


class GOESProductLocatorABIPM(GOESProductLocatorABINSO):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Multi-band primary ABI products.
    """

    # Available single-band products from the GOES-R Series's ABI
    # instrument:
    AVAILABLE_PRODUCTS: dict[str, str] = {
        "MCMIP": "Multi-band Cloud and Moisture Imagery Product",
    }

    def __init__(self, name: str, scene: str, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI derived product
        locator object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        self._validate_product(name, self.AVAILABLE_PRODUCTS)

        super().__init__(
            name=name,
            scene=scene,
            origin=origin,
        )


class GOESProductLocatorMCMIP(GOESProductLocatorABIPM):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Cloud and Moisture Imagery Product (CMIP).
    """

    PRODUCT_NAME: str = "MCMIP"

    def __init__(self, scene: str, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI multi-band
        primary product locator object.

        Parameters
        ----------
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        super().__init__(name=self.PRODUCT_NAME, scene=scene, origin=origin)



"""
GOES Imager Projection data extraction.

This module provides classes to extract and represent GOES Imager
Projection information from GOES satellite netCDF data files.

Classes
-------
GOESOrbitGeometry
    Represent GOES-R series satellite orbit geometry information.
GOESGlobe
    Represent GOES-R series satellite globe definition.
GOESABIFixedGrid
    Represent GOES-R series satellite ABI Fixed Grid projection data.
GOESProjection
    Represent GOES-R series satellite projection information.
"""

from numpy import float64

from ..netcdf import DatasetView, data, variable
from ..utils.array import ArrayFloat32, ArrayFloat64

imager_proj = variable("goes_imager_projection")


class GOESOrbitGeometry(DatasetView):
    """
    Represent GOES-R series satellite orbit geometry information.

    Note
    ----
    For information on GOES orbit geometry, see [1]_ and Section 4.2. of
    [2]_

    Attributes
    ----------
    longitude_of_projection_origin : np.float64
        The longitude of the ideal satellite sub-point in East degrees.
    perspective_point_height : np.float64
        The height of the perspective point, or the height above the
        Earth's surface the ideal satellite sub-point, in meters.
    sweep_angle_axis : str
        The sweep angle axis.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """

    # Information about the projection
    longitude_of_projection_origin: float64 = imager_proj.attribute()
    perspective_point_height: float64 = imager_proj.attribute()
    sweep_angle_axis: str = imager_proj.attribute()


class GOESGlobe(DatasetView):
    """
    Represent GOES-R series satellite globe definition.

    Note
    ----
    For information on GOES orbit geometry and Earth globe definition,
    see [1]_ and Section 4.2. of [2]_

    Attributes
    ----------
    semi_major_axis : np.float64
        The semi-major axis of the globe in meters.
    semi_minor_axis : np.float64
        The semi-minor axis of the globe in meters.
    inverse_flattening : np.float64
        The inverse flattening of the globe.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """

    # Information about the globe
    semi_major_axis: float64 = imager_proj.attribute()
    semi_minor_axis: float64 = imager_proj.attribute()
    inverse_flattening: float64 = imager_proj.attribute()


class GOESABIFixedGrid(DatasetView):
    """
    Represent GOES-R series satellite ABI Fixed Grid projection data.

    The GOES Imager Projection, also called the ABI Fixed Grid, is the
    projection information included in all ABI Level 1b radiance data
    files and most ABI Level 2 derived product data files. It is a map
    projection based on the geostationary viewing perspective of the
    GOES-East or GOES-West satellite.

    Notes
    -----
    For information on GOES Imager Projection and GOES orbit geometry,
    see [1]_ and Section 4.2. of [2]_

    Attributes
    ----------
    x : ArrayFloat32
        1D array of E/W scanning angles in radians.
    y : ArrayFloat32
        1D array of N/S elevation angles in radians.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """

    x: ArrayFloat32 = data()
    y: ArrayFloat32 = data()


def to_float64(array: ArrayFloat32) -> ArrayFloat64:
    """
    Convert a float32 array to a float64 array.

    Parameters
    ----------
    array : ArrayFloat32
        The float32 array to convert.

    Returns
    -------
    ArrayFloat64
        The float64 array.
    """
    return array.astype(float64)


class GOESProjection(GOESOrbitGeometry, GOESGlobe):
    """
    Represent GOES-R series satellite projection information.

    The GOES Imager Projection, also called ABI Fixed Grid Projection,
    is a map projection relative to the GOES satellite point of view.
    Units: latitude in °N (°S < 0), longitude in °E (°W < 0)

    Note
    ----
    For information on GOES Imager Projection, see [1]_ and Section
    4.2.8 of [2]_

    Attributes
    ----------
    x : ArrayFloat64
        1D array of E/W scanning angles in radians.
    y : ArrayFloat64
        1D array of N/S elevation angles in radians.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """

    # Information about the fixed grid
    x: ArrayFloat64 = data(convert=to_float64)
    y: ArrayFloat64 = data(convert=to_float64)



from numpy import float64


class GlobeParameters:
    """
    A class to store the globe parameters.

    Attributes
    ----------
    semi_major_axis : float64
        The semi-major axis of the ellipsoid in meters.
    semi_minor_axis : float64
        The semi-minor axis of the ellipsoid in meters.
    inverse_flattening : float64
        The inverse flattening of the ellipsoid.
    """

    semi_major_axis: float64
    semi_minor_axis: float64
    inverse_flattening: float64

    def __init__(
        self,
        semi_major_axis: float64,
        semi_minor_axis: float64,
        inverse_flattening: float64,
    ) -> None:
        """
        Initialize the GlobeParameters object.

        Parameters
        ----------
        semi_major_axis : float64
            The semi-major axis of the ellipsoid in meters.
        semi_minor_axis : float64
            The semi-minor axis of the ellipsoid in meters.
        inverse_flattening : float64
            The inverse flattening of the ellipsoid.
        """
        self.semi_major_axis = semi_major_axis
        self.semi_minor_axis = semi_minor_axis
        self.inverse_flattening = inverse_flattening

    @property
    def flattening(self) -> float64:
        """
        Get the flattening of the globe.

        Returns
        -------
        np.float64
            The flattening of the globe.
        """
        return 1.0 / self.inverse_flattening



"""
Provide functions for latitude and longitude grid calculation.

These functions calculate latitude and longitude grid data from GOES ABI
fixed grid projection data. GOES ABI fixed grid projection is a map
projection relative to the GOES satellite.

Notes
-----
See [2]_, GOES-R Product User Guide (PUG) Volume 5 (L2 products) Section
4.2.8 for details & example of calculations.

Classes
-------
GeostationaryParameters
    Class to store GOES ABI fixed grid projection parameters.

Functions
---------
geos_to_latlon_grid
    Calculate latitude and longitude grids using a given algorithm.
geos_to_latlon_grid_cartopy
    Calculate latitude and longitude grids using the cartopy package.
geos_to_latlon_grid_goesdl
    Calculate latitude and longitude grids using an optimized version of
    classic's algorithm.
geos_to_latlon_grid_pyproj
    Calculate latitude and longitude grids using the pyproj package.
calculate_pixel_edges
    Calculate the pixel edges of the GOES ABI fixed grid.

References
----------
.. [1] STAR Atmospheric Composition Product Training, "GOES Imager
    Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
    https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
.. [2] Aerosols and Atmospheric Composition Science Team, "Python Short
    Demo: Calculate Latitude and Longitude from GOES Imager Projection
    (ABI Fixed Grid) Information", NOAA/NESDIS/STAR, 2024.
    https://www.star.nesdis.noaa.gov/atmospheric-composition-training/python_abi_lat_lon.php
.. [3] GOES-R, " GOES-R Series Product Definition and User’s Guide
    (PUG), Volume 5: Level 2+ Products", Version 2.4, NASA/NOAA/NESDIS,
    2022.
    https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf

"""

from .geos_to_latlon_grid import geos_to_latlon_grid
from .geos_to_latlon_grid_cartopy import geos_to_latlon_grid_cartopy
from .geos_to_latlon_grid_goesdl import geos_to_latlon_grid_goesdl
from .geos_to_latlon_grid_pyproj import geos_to_latlon_grid_pyproj
from .geostationary_parameters import GeostationaryParameters
from .helpers import calculate_pixel_edges
from .rectangular_region import RectangularRegion

__all__ = [
    "calculate_pixel_edges",
    "geos_to_latlon_grid",
    "geos_to_latlon_grid_cartopy",
    "geos_to_latlon_grid_goesdl",
    "geos_to_latlon_grid_pyproj",
    "GeostationaryParameters",
    "RectangularRegion",
]



from typing import Any, cast

from netCDF4 import Dataset  # pylint: disable=no-name-in-module
from numpy import nan

from ..geodesy import RectangularRegion
from ..netcdf import DatasetView, HasStrHelp, variable
from ..utils.array import ArrayBool, ArrayFloat32, MaskedFloat32
from .metadata import MeasurementMetadata
from .netcdf_geodetic import GSLatLonGrid, LimitType


class GSImageData(HasStrHelp):

    raster: MaskedFloat32


class GSImage(HasStrHelp):

    grid: GSLatLonGrid
    raster: MaskedFloat32

    metadata: MeasurementMetadata

    def __init__(
        self, record: Dataset, channel: str, grid: GSLatLonGrid
    ) -> None:
        data = self._extract_image(
            record, channel, grid.lon_limits, grid.lat_limits
        )

        self.grid = grid
        self.raster = data.raster

        self.metadata = self._extract_metadata(record, channel)

    @staticmethod
    def _extract_image(
        record: Dataset,
        channel: str,
        lon_limits: LimitType,
        lat_limits: LimitType,
    ) -> "GSImageData":
        def slice(x: Any) -> Any:
            min_lon, max_lon = lon_limits
            min_lat, max_lat = lat_limits
            return x[0, min_lat:max_lat, min_lon:max_lon]

        class _GSImageData(DatasetView):
            raster: MaskedFloat32 = variable(channel).array(filter=slice)

        data = _GSImageData(record)

        data.raster.data[data.raster.mask] = nan

        return cast(GSImageData, data)

    @staticmethod
    def _extract_metadata(record: Dataset, name: str) -> MeasurementMetadata:
        measurement = variable(name)

        class _ImageMetata(DatasetView):
            long_name: str = measurement.attribute()
            standard_name: str = measurement.attribute()
            units: str = measurement.attribute()
            content_type: str = measurement.attribute("coverage_content_type")
            coordinates: str = measurement.attribute()
            range: ArrayFloat32 = measurement.attribute("actual_range")
            shape: tuple[int] = measurement.attribute()

        metadata = _ImageMetata(record)

        return MeasurementMetadata(metadata)

    @property
    def image(self) -> ArrayFloat32:
        return cast(ArrayFloat32, self.raster.data)

    @property
    def mask(self) -> ArrayBool:
        return cast(ArrayBool, self.raster.mask)

    @property
    def region(self) -> RectangularRegion:
        return self.grid.region



import numpy as np


def interpolate_value(x, xp, fp):
    """
    Realiza una interpolación lineal utilizando np.interp y, si el punto
    está fuera del dominio, realiza una extrapolación lineal.
    """
    if x < xp[0]:
        # Extrapolación lineal a la izquierda
        slope = (fp[1] - fp[0]) / (xp[1] - xp[0])
        return fp[0] + slope * (x - xp[0])

    if x > xp[-1]:
        # Extrapolación lineal a la derecha
        slope = (fp[-1] - fp[-2]) / (xp[-1] - xp[-2])
        return fp[-1] + slope * (x - xp[-1])

    # Interpolación dentro del dominio
    return np.interp(x, xp, fp)


def interpolate_coordinates(t, timestamps, longitudes, latitudes):
    """Interpola o extrapola las coordenadas para un timestamp dado."""
    # Convertir a arrays NumPy si aún no lo están
    timestamps = np.array(timestamps)
    latitudes = np.array(latitudes)
    longitudes = np.array(longitudes)

    # Interpolar o extrapolar
    lon_interp = interpolate_value(t, timestamps, longitudes)
    lat_interp = interpolate_value(t, timestamps, latitudes)

    return lon_interp, lat_interp



"""
Provide locator for GOES-R Series imagery dataset's GLM products.

Classes:
    - GOESProductLocatorGLM: All Geostationary Lightning Mapper (GLM)
      products.
"""

from .locator import GOESProductLocator


class GOESProductLocatorGLM(GOESProductLocator):
    """
    Product locator for GOES-R Series imagery dataset's GLM products.

    Instrument: Geostationary Lightning Mapper (GLM).
    Product: All GLM products.
    """

    AVAILABLE_PRODUCTS: dict[str, str] = {
        "LCFA": "Lightning Cluster-Filter Algorithm"
    }

    INSTRUMENT_NAME: str = "GLM"
    PRODUCT_LEVEL: str = "L2"

    def __init__(self, name: str, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset GLM product locator.

        Constructs a GOES-R Series imagery dataset GLM product locator
        object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset GLM product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        origin : str
            The origin of the GOES-R Series imagery dataset GLM product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.

        Notes
        -----
        ValueError is raised if the provided product name is invalid.
        """
        self._validate_product(name, self.AVAILABLE_PRODUCTS)

        super().__init__(
            name=name,
            level=self.PRODUCT_LEVEL,
            scene="",
            instrument=self.INSTRUMENT_NAME,
            modes=[],
            channels=[],
            origin=origin,
        )



import matplotlib.pyplot as plt


def plot_difference(
    matrices,
    filename=None,
    captions=None,
    cmap="viridis",
    figsize=(8, 8),
    figdpi=200,
    imgdpi=200,
    title=None,
    subtitle=None,
    show=True,
):
    """
    Plots four matrices (interpreted as images) in two rows.
    
    Args:
        titles: An optional list of three strings, specifying the titles for each plot.
            Defaults to None (no titles).
        cmap: The colormap to use for displaying the matrices. Defaults to 'viridis'.
            See matplotlib documentation for available colormaps.
    """
    if isinstance(cmap, tuple):
        cmap = list(cmap)

    if isinstance(cmap, str):
        cmap = [cmap] * 4
    elif len(cmap) == 1:
        cmap = cmap * 4
    elif len(cmap) == 2:
        cmap = [cmap[0]] * 3 + [cmap[1]] * 2
    elif len(cmap) == 3:
        cmap = [cmap[0]] * 2 + cmap[1:]

    fig, axes = plt.subplots(2, 2, figsize=figsize, dpi=figdpi)

    if captions is None:
        captions = []
    elif isinstance(captions, str):
        captions = [captions]

    title_size = 7
    subtitle_size = 4.5
    caption_size = 4.5
    label_size = 4
    tick_size = 3.5
    axis_size = 4
    y_title = 0.93

    if subtitle:
        y_title = 0.95
        fig.text(0.5, 0.903, subtitle, ha='center', va='bottom', fontsize=subtitle_size)

    if title:
        fig.suptitle(title, fontsize=title_size, y=y_title)

    if len(captions) == 1:
        fig.text(0.5, 0.92, captions[0], ha='center', va='bottom', fontsize=caption_size)
    elif len(captions) == 2:
        fig.text(0.5, 0.92, captions[0], ha='center', va='bottom', fontsize=caption_size)
        fig.text(0.5, 0.50, captions[1], ha='center', va='bottom', fontsize=caption_size)
    elif len(captions) == 3:
        fig.text(0.5, 0.92, captions[0], ha='center', va='bottom', fontsize=caption_size)
        axes[1, 0].set_title(captions[1], fontsize=caption_size)
        axes[1, 1].set_title(captions[2], fontsize=caption_size)
    elif len(captions) == 4:
        axes[0, 0].set_title(captions[0], fontsize=caption_size)
        axes[0, 1].set_title(captions[1], fontsize=caption_size)
        axes[1, 0].set_title(captions[2], fontsize=caption_size)
        axes[1, 1].set_title(captions[3], fontsize=caption_size)

    plt.subplots_adjust(right=0.85, hspace=0.3, wspace=1.0)

    vmin = (200, 200, -90, 0)
    vmax = (300, 300, +90, 1)

    # Plot the matrices
    for i in range(2):
        for j in range(2):
            k = i * 2 + j
            ax = axes[i, j]
            im = ax.imshow(matrices[k], cmap=cmap[k], vmin=vmin[k], vmax=vmax[k])

            ax.tick_params(
                left=True,
                right=False,
                bottom=True,
                top=False,
                labelleft=True,
                labelright=False,
                labelbottom=True,
                labeltop=False,
                length=1.1,
                width=0.3,
                labelsize=tick_size,
                pad=1.1,
                labelcolor="black",
            )
            ax.set_xlabel("Pixel indices", color="black", fontsize=axis_size, labelpad=1.0)
            ax.set_ylabel("Pixel indices", color="black", fontsize=axis_size, labelpad=0.5)

            [x.set_linewidth(0.3) for x in ax.spines.values()]

            cb_width = 0.05
            cb_left = 1.07
            cb_bottom = 0
            cb_height = 1.0

            # Crear un nuevo eje para la barra de color
            cax = ax.inset_axes([cb_left, cb_bottom, cb_width, cb_height])

            # Añadir la barra de color al nuevo eje
            cb = fig.colorbar(im, cax=cax)
            cax.tick_params(
                length=1.1,
                width=0.3,
                labelsize=label_size,
            )
            # cb.set_label(label="K", size=label_size, color="black", weight="normal")
            
            [x.set_linewidth(0.3) for x in cax.spines.values()]

    # Adjust layout to prevent overlapping titles
    plt.tight_layout()
    
    # Show the media file
    if filename:
        plt.savefig(filename, dpi=imgdpi, bbox_inches='tight')
    
    # Show the plot
    if show:
        plt.show()
    else:
        plt.close(fig)



from numpy import float64

from ..utils.array import ArrayFloat64
from .geosproj_parameters import GeosProjParameters
from .globe_parameters import GlobeParameters


class GeostationaryParameters:
    """
    A class to store the geostationary grid and projection parameters.

    Attributes
    ----------
    orbit : GeostationaryParameters
        The parameters of the geostationary orbit.
    globe : GlobeParameters
        The parameters of the globe.
    x : ArrayFloat64
        The x-coordinate grid in radians.
    y : ArrayFloat64
        The y-coordinate grid in radians.
    """

    # Information about the projection
    orbit: GeosProjParameters

    # Information about the globe
    globe: GlobeParameters

    # Information about the fixed grid
    x: ArrayFloat64
    y: ArrayFloat64

    def __init__(
        self,
        orbit: GeosProjParameters,
        globe: GlobeParameters,
        xy_grid: tuple[ArrayFloat64, ArrayFloat64],
    ) -> None:
        """
        Initialize the class.

        Parameters
        ----------
        orbit : GeostationaryParameters
            The parameters of the geostationary orbit.
        globe : GlobeParameters
            The parameters of the globe.
        xy_grid : tuple[ArrayFloat64, ArrayFloat64]
            The x and y grid data in radians.
        """
        self.orbit = orbit
        self.globe = globe
        self.x = xy_grid[0]
        self.y = xy_grid[1]

    @property
    def orbital_radius(self) -> float64:
        """
        Calculate the orbital radius of the GOES satellite in meters.

        Returns
        -------
        np.float64
            The orbital radius of the GOES satellite in meters.
        """
        return self.orbit.perspective_point_height + self.globe.semi_major_axis

    @property
    def x_m(self) -> ArrayFloat64:
        """
        Calculate the x-coordinate fixed grid in meters.

        Returns
        -------
        ArrayFloat64
            The x-coordinate fixed grid in meters.
        """
        return self.orbit.perspective_point_height * self.x

    @property
    def y_m(self) -> ArrayFloat64:
        """
        Calculate the y-coordinate fixed grid in meters.

        Returns
        -------
        ArrayFloat64
            The y-coordinate fixed grid in meters.
        """
        return self.orbit.perspective_point_height * self.y



from typing import TextIO


def str_to_float(value_strings):
    return [float(value) for value in value_strings]


def get_track_data(lines: list[str]):
    clines = []
    for line in lines:
        cline = []
        for i, word in enumerate(line.split(",")):
            if i in {0, 1, 4, 5}:
                cline.append(word.strip())
        line = []
        for i in range(len(cline)):
            if i == 0:
                cline[i] = f"{cline[i][:4]}-{cline[i][4:6]}-{cline[i][6:8]}"
            if i == 1:
                cline[i] = f"{int(cline[i]):0>4}"
                cline[i] = f"T{cline[i][:2]}:{cline[i][2:]}Z"
            if i == 2:
                cline[i] = (
                    f"+{cline[i][:-1]}"
                    if cline[i][-1] == "N"
                    else f"-{cline[i][:-1]}"
                )
            if i == 3:
                cline[i] = (
                    f"+{cline[i][:-1]}"
                    if cline[i][-1] == "E"
                    else f"-{cline[i][:-1]}"
                )
        line = [cline[0] + cline[1], cline[2], cline[3]]
        clines.append(line)

    iso_dates, latitudes, longitudes = zip(*clines)

    return (
        iso_to_timestamp(iso_dates),
        str_to_float(latitudes),
        str_to_float(longitudes),
    )


def parse_header(line: list[str]):
    parts = line.strip().split(",")

    identifier = parts[0].strip()
    sector = identifier[:2]
    number = int(identifier[2:4])
    year = int(identifier[4:8])

    name = parts[1].strip()
    nlines = int(parts[2].strip())

    return year, name, nlines, sector, number


def read_lines(file: TextIO, nlines: int):
    data_lines = []
    for _ in range(nlines):
        try:
            data_lines.append(next(file))
        except StopIteration:
            break
    return data_lines


def skip_lines(file: TextIO, nlines: int):
    for _ in range(nlines):
        try:
            next(file)
        except StopIteration:
            break


def parse_file(file: TextIO, target_name: str, target_year: str):
    while True:
        try:
            line = next(file)
            try:
                year, name, nlines, sector, number = parse_header(line)
                if year == target_year and name == target_name:
                    return read_lines(file, nlines)
                else:
                    skip_lines(file, nlines)
            except ValueError:
                continue
            except IndexError:
                continue
        except StopIteration:
            break
    return []


def parse_hurdat(
    file_path: str, target_name: str, target_year: str
) -> list[str]:
    target_name = target_name.upper()
    try:
        with open(file_path, "r") as file:
            lines = parse_file(file, target_name, target_year)
            if lines:
                return lines
        return []
    except FileNotFoundError:
        print(f"Error: El archivo no se encontró en la ruta: {file_path}")
        return []
    except Exception as e:
        print(f"Ocurrió un error: {e}")
        return []



from pathlib import Path
from typing import TextIO

from .clr_table import clr_utility
from .constants import CLR_MAX, CM_BGR, CM_RGB, UNNAMED_TABLE
from .shared import (
    ColorEntry,
    DomainData,
    PaletteData,
    PaletteItem,
    RGBValue,
    ValueTables,
)

MCIDAS_EU_SIGNATURE = "EU TABLE"
MCIDAS_EU_KEYWORD = (
    "EU",
    "TABLE",
    "Brightness",
    "Red",
    "Green",
    "Blue",
    "min",
    "max",
    "---",
    "------",
)


class eu_utility(clr_utility):

    @staticmethod
    def add_color_table_header(lines: list[str], name: str, rgb: bool) -> None:
        lines.append(f"{MCIDAS_EU_SIGNATURE} {name.upper()}")
        i, j, k = (3, 4, 5) if rgb else (5, 4, 3)
        lines.append(
            f" {MCIDAS_EU_KEYWORD[2]}"
            f"{MCIDAS_EU_KEYWORD[i]:>6}"
            f"{MCIDAS_EU_KEYWORD[j]:>11}"
            f"{MCIDAS_EU_KEYWORD[k]:>9}"
        )
        hdr = f"{MCIDAS_EU_KEYWORD[6]} {MCIDAS_EU_KEYWORD[7]}"
        lines.append(f"{hdr:>9}{hdr:>10}{hdr:>10}{hdr:>10}")
        hdr = f"{MCIDAS_EU_KEYWORD[8]} {MCIDAS_EU_KEYWORD[8]}"
        lines.append(f"{hdr:>9}{hdr:>10}{hdr:>10}{hdr:>10}")

    @classmethod
    def create_file(
        cls,
        path: str | Path,
        name: str,
        table: PaletteData,
        extent: DomainData,
        rgb: bool = False,
    ) -> None:
        """
        Create a file with the color table.

        Parameters
        ----------
        path : str or Path
            Path to the file where the color table will be saved.
        name : str
            The name of the color table.
        table : PaletteTable
            The color table data.
        extent : DomainData
            The palette extent containing the minimum and maximum
            defined values.
        rgb : bool, optional
            Flag indicating if the color model is RGB, by default False.
        """
        lines: list[str] = []

        eu_utility.add_color_table_header(lines, name, rgb)
        cls._create_color_table(lines, table, extent, rgb)

        with open(path, "w", encoding="utf-8", newline="\n") as file:
            cls._write_color_table_file(file, lines)

    @staticmethod
    def is_eu_table(header: str) -> bool:
        return MCIDAS_EU_SIGNATURE in header

    @classmethod
    def _create_color_table(
        cls,
        lines: list[str],
        table: PaletteData,
        extent: DomainData,
        rgb: bool,
    ) -> None:
        x_min, x_max = extent
        x_range = x_max - x_min

        if rgb:
            table = [(x, r, g, b) for x, b, g, r in table]

        for i in range(0, len(table), 2):
            x_lo, b_lo, g_lo, r_lo = cls._get_color_entry(
                table[i],
                x_min,
                x_range,
            )

            x_hi, b_hi, g_hi, r_hi = cls._get_color_entry(
                table[i + 1],
                x_min,
                x_range,
            )

            line = (
                f"{x_lo:>5}{x_hi:>4}{b_lo:>6}{b_hi:>4}"
                f"{g_lo:>6}{g_hi:>4}{r_lo:>6}{r_hi:>4}"
            )

            lines.append(line)

    @classmethod
    def _get_color_entry(
        cls, entry: ColorEntry, x_min: float, x_range: float
    ) -> ColorEntry:
        x_lo, b_lo, g_lo, r_lo = entry

        x_lo = round(x_min + x_lo * x_range)
        b_lo = round(b_lo * CLR_MAX)
        g_lo = round(g_lo * CLR_MAX)
        r_lo = round(r_lo * CLR_MAX)

        return x_lo, b_lo, g_lo, r_lo

    @classmethod
    def parse_eu_table(
        cls, lines: list[str]
    ) -> tuple[PaletteItem, str, DomainData]:
        j: list[float] = []
        b: list[float] = []
        g: list[float] = []
        r: list[float] = []

        color_model = CM_BGR

        for line in lines:
            # Split line into list of strings of keywords or values
            ls = line.split()

            # Check for alternative color model
            if ls[0] == MCIDAS_EU_KEYWORD[2] and ls[1] == MCIDAS_EU_KEYWORD[3]:
                color_model = CM_RGB

            # Ignore header lines
            if ls[0] in MCIDAS_EU_KEYWORD:
                continue

            j.extend((float(ls[0]), float(ls[1])))
            b.extend((float(ls[2]), float(ls[3])))
            g.extend((float(ls[4]), float(ls[5])))
            r.extend((float(ls[6]), float(ls[7])))

        bg = b[0], g[0], r[0]
        fg = b[-1], g[-1], r[-1]
        nn = 1.0, 0.0, 1.0

        b, g, r = cls._process_eu_colors(color_model, b, g, r)

        extent = j[0], j[-1]

        entries = cls._make_color_entries(j, b, g, r)
        stock = cls._process_eu_stock(color_model, bg, fg, nn)

        name = UNNAMED_TABLE
        if len(lines[0]) > len(MCIDAS_EU_SIGNATURE):
            name = lines[0][len(MCIDAS_EU_SIGNATURE) + 1 :]
            name = name.strip()

        return (entries, stock), name, extent

    @classmethod
    def _process_eu_colors(
        cls,
        color_model: str,
        b: list[float],
        g: list[float],
        r: list[float],
    ) -> ValueTables:
        # Normalize color values
        b = list(map(cls._normalize_color, b))
        g = list(map(cls._normalize_color, g))
        r = list(map(cls._normalize_color, r))

        # Convert color model if necessary
        if color_model == CM_RGB:
            r, b = b, r

        return b, g, r

    @classmethod
    def _process_eu_stock(
        cls,
        color_model: str,
        bg: RGBValue,
        fg: RGBValue,
        nn: RGBValue,
    ) -> list[RGBValue]:
        packed = (bg, fg, nn)
        u, v, w = zip(*packed)
        b, g, r = list(u), list(v), list(w)

        b, g, r = cls._process_eu_colors(color_model, b, g, r)

        return list(zip(r, g, b))

    @staticmethod
    def _write_color_table_file(file: TextIO, lines: list[str]) -> None:
        for line in lines:
            file.write(f"{line:<85}\n")



from typing import Any

from numpy import empty, float32, int8

from ..netcdf import HasStrHelp
from ..utils.array import ArrayFloat32, ArrayInt8
from .constants import NA


class VariableMetadata(HasStrHelp):
    long_name: str
    standard_name: str
    comment: str
    units: str
    shape: tuple[int, ...]

    def __init__(self, metadata: Any) -> None:
        self.long_name = getattr(metadata, "long_name", NA)
        self.standard_name = getattr(metadata, "standard_name", NA)
        self.comment = getattr(metadata, "comment", NA)
        self.units = getattr(metadata, "units", NA)
        self.shape = getattr(metadata, "shape", (0,))


class ContentMetadata(VariableMetadata):
    content_type: str

    def __init__(self, metadata: Any) -> None:
        super().__init__(metadata)

        self.content_type = getattr(metadata, "content_type", NA)


class CoordinateMetadata(ContentMetadata):
    axis: str

    def __init__(self, metadata: Any) -> None:
        super().__init__(metadata)

        self.axis = getattr(metadata, "axis", NA)


class FramedMetadata(ContentMetadata):
    coordinates: str

    def __init__(self, metadata: Any) -> None:
        super().__init__(metadata)

        self.coordinates = getattr(metadata, "coordinates", NA)


class MeasurementMetadata(FramedMetadata):
    range: ArrayFloat32

    def __init__(self, metadata: Any) -> None:
        super().__init__(metadata)

        self.range = getattr(metadata, "range", empty((0,), dtype=float32))


class TimeMetadata(CoordinateMetadata):
    calendar: str

    def __init__(self, metadata: Any) -> None:
        super().__init__(metadata)

        self.calendar = getattr(metadata, "calendar", NA)


class DeltaTimeMetadata(FramedMetadata):
    range: ArrayInt8

    def __init__(self, metadata: Any) -> None:
        super().__init__(metadata)

        self.range = getattr(metadata, "range", empty((0,), dtype=int8))



"""
Utilities for generating string representation of classes.

Provide helper functions for generating string representations of class
instances and their attributes, particularly for classes that inherit
from a specified root class. It includes functions to retrieve attribute
names, format attribute values, and handle NumPy arrays.

Functions:
    help_str(this: Any, root: type, indent: str = "    ") -> str:
        Generate a formatted string representation of an instance,
        including its class name and attribute values, with indentation
        for readability.
"""

from typing import Any

from numpy import float32, ndarray
from numpy.typing import NDArray


def _get_attribute_names(this: Any, root: type) -> list[str]:
    # Retrieve a list of attribute names for a given instance and its
    # superclasses up to a specified root class.
    __dict__: dict[str, Any] = getattr(this, "__dict__", {})
    all_attributes: list[str] = list(__dict__.keys())
    mro_classes: tuple[type, ...] = this.__class__.__mro__
    for cls in mro_classes:
        if not hasattr(cls, "__annotations__"):
            continue
        if issubclass(cls, root) and cls != root:
            all_attributes.extend(
                [
                    key
                    for key in cls.__annotations__.keys()
                    if key not in all_attributes
                ]
            )
    return all_attributes


def _get_attribute_value(attribute_value: Any, root: type, indent: str) -> str:
    # Format the value of an attribute as a string, handling special
    # cases such as None, NumPy arrays, floats, strings, booleans, and
    # instances of the root class.
    if attribute_value is None:
        return "'not available'"
    if isinstance(attribute_value, ndarray):
        return _str_ndarray(attribute_value)
    if isinstance(attribute_value, float):
        return f"{attribute_value:.6f}"
    if isinstance(attribute_value, str):
        return f"'{attribute_value}'"
    if isinstance(attribute_value, bool):
        return f"{attribute_value}".lower()
    if isinstance(attribute_value, root):
        return help_str(attribute_value, root, f"{indent}    ")
    return str(attribute_value)


def _str_ndarray(array: NDArray[float32]) -> str:
    # Generates a string representation of a NumPy array, including its
    # shape and data type. Returns a string representation of a NumPy
    # array, including its type, shape, and data type.
    shape: str = f"({array.size})" if array.ndim == 1 else f"{array.shape}"
    dtype = f"{array.dtype}"
    value = f"{type(array)}, shape={shape}, dtype={dtype}>"
    return value.replace(">,", ",")


def help_str(this: Any, root: type = object, indent: str = "    ") -> str:
    """
    Generate a formatted string representation of an instance.

    Generate a formatted string representation of an instance, including
    its class name and attribute values, with indentation for
    readability.

    Parameters
    ----------
    this : Any
        An instance of a class.
    root : type, optional
        The root class from which to extract attributes.
    indent : str, optional
        The indentation string, by default four spaces.

    Returns
    -------
    str
        A formatted string representation of the instance.
    """
    attributes = [str(this.__class__)]
    all_attributes = _get_attribute_names(this, root)
    for attribute_name in all_attributes:
        attribute_value = getattr(this, attribute_name, None)
        attribute_value = _get_attribute_value(attribute_value, root, indent)
        attributes.append(f"{indent}{attribute_name}: {attribute_value}")
    for attribute_name in dir(this.__class__):
        attribute_value = getattr(this.__class__, attribute_name)
        if isinstance(attribute_value, property):
            attribute_value = getattr(this, attribute_name, None)
            attribute_value = _get_attribute_value(
                attribute_value, root, indent
            )
            attributes.append(f"{indent}{attribute_name}: {attribute_value}")
    return "\n".join(attributes)


class HasStrHelp:
    """
    Provide a string representation of its instance.

    Methods
    -------
    __str__() -> str
        Returns a string representation of the object's class instance.
    """

    def __str__(self) -> str:
        # Returns a representation of the object's class instance.
        return help_str(self, HasStrHelp)



import matplotlib.pyplot as plt
import numpy as np

from .scale import EnhancementScale
from .ticks import ColorbarTicks


def show_colormap(
    scale: EnhancementScale,
    measurement: str = "",
    offset: float = 0.0,
    nticks: int = 14,
) -> None:
    # Example data
    vmin, vmax = scale.domain
    vmin, vmax = vmin + offset, vmax + offset
    data = np.linspace(vmin, vmax, scale.ncolors)[None, :]

    # Color scale map preview plot
    fig, ax = plt.subplots()

    # Plot the example data
    tmin, tmax = scale.extent
    tmin, tmax = tmin + offset, tmax + offset
    im = ax.imshow(data, aspect="auto", cmap=scale.cmap, vmin=tmin, vmax=tmax)

    # Create the color bar with the measurement scale
    cbar = fig.colorbar(im, orientation="horizontal")

    cbar_ref = ColorbarTicks((tmin, tmax), nticks, 0)
    cbar.set_ticks(cbar_ref.cticks)

    # Note: Alternatively `norm=scale.cnorm` can be passed to imshow and
    # `cbar.ax.minorticks_off()` is used to inhibit the minor ticks.

    # Hide the y-axis since it does not make sense
    ax.tick_params(
        axis="y", which="both", left=False, right=False, labelleft=False
    )

    # Set the title, x-axis label and the colorbar caption
    plt.title(scale.name)

    ax.set_xlabel("Color Index", color="black", labelpad=3.0)

    if not measurement:
        measurement = "Measurement"

    cbar.set_label(label=measurement, color="black", weight="normal")

    plt.show()



from .constants import CLR_MAX
from .shared import ColorEntry


class clr_utility:

    @classmethod
    def _make_color_entries(
        cls, j: list[float], b: list[float], g: list[float], r: list[float]
    ) -> list[ColorEntry]:
        length = j[-1] - j[0]
        x = [(k - j[0]) / length for k in j]

        return list(zip(x, b, g, r))

    @staticmethod
    def _normalize_color(x: float) -> float:
        return x / CLR_MAX



import unittest
from datetime import datetime, timedelta, timezone

from goesdl.gridsat import GridSatProductLocatorGC


class TestPassed(Exception):
    """
    Exception raised when a test has passed.
    """


class TestGridSatProductLocatorGC(unittest.TestCase):
    # This set of tests covers the initialisation of an object of
    # the `GridSatProductLocatorGC` class, including the handling of
    # invalid parameters. It also tests the `ProductLocator` interface
    # implemented by the `GridSatProductLocatorGC` and parent classes.

    def setUp(self) -> None:
        # Initialise the locator with a valid scene, and all supported
        # origins and versions.
        self.locator: GridSatProductLocatorGC = GridSatProductLocatorGC(
            scene=self._supported_scenes()[0],
            origins=self._supported_origins(),
            versions=self._supported_versions(),
        )
        self.longMessage = True

    def setUpLocator(
        self, scene: str = "", origin: str = "", version: str = ""
    ) -> None:
        self.locator: GridSatProductLocatorGC = GridSatProductLocatorGC(
            scene=scene or self._supported_scenes()[0],
            origins=origin or self._supported_origins(),
            versions=version or self._supported_versions(),
        )

    # ------------------------------------------------------------------
    # A minimal subset of tests are designed to reach 100% code coverage
    # of the `GridSatProductLocatorGC` class hierarchy. The remaining
    # complementary tests help to reach 100% test coverage.
    #
    # Note: For completeness, `match()` is tested against valid filename
    # format with an unknown version, and against a valid filename with
    # an invalid date.
    #
    # Although date conversion is performed under the hood by the
    # framework and `get_datetime()` does not perform any checking for
    # invalid dates, it is tested here against a valid filename with an
    # invalid date to ensure that the method is consistent. It is also
    # tested against a `datetime` object set in a non-UTC timezone.

    def test_init_supported_scenes(self) -> None:
        for scene in self._supported_scenes():
            with self.subTest(scene=scene):
                with self.assertRaises(TestPassed):
                    GridSatProductLocatorGC(
                        scene=scene,
                        origins=self._supported_origins(),
                        versions=self._supported_versions(),
                    )
                    raise TestPassed()

    def test_init_unsupported_scenes(self) -> None:
        for scene in self._unsupported_scenes():
            with self.subTest(scene=scene):
                with self.assertRaises(ValueError):
                    GridSatProductLocatorGC(
                        scene=scene,
                        origins=self._supported_origins(),
                        versions=self._supported_versions(),
                    )

    def test_init_supported_origins(self) -> None:
        for origin in self._supported_origins():
            with self.subTest(origin=origin):
                with self.assertRaises(TestPassed):
                    GridSatProductLocatorGC(
                        scene=self._supported_scenes()[0],
                        origins=origin,
                        versions=self._supported_versions(),
                    )
                    raise TestPassed()

    def test_init_unsupported_origins(self) -> None:
        for origin in self._unsupported_origins():
            with self.subTest(origin=origin):
                with self.assertRaises(ValueError):
                    GridSatProductLocatorGC(
                        scene=self._supported_scenes()[0],
                        origins=origin,
                        versions=self._supported_versions(),
                    )

    def test_init_default_version(self) -> None:
        # Initialise the locator with the default version.
        with self.assertRaises(TestPassed):
            GridSatProductLocatorGC(
                scene=self._supported_scenes()[0],
                origins=self._supported_origins()[0],
            )
            raise TestPassed()

    def test_init_supported_versions(self) -> None:
        for version in self._supported_versions():
            with self.subTest(version=version):
                with self.assertRaises(TestPassed):
                    GridSatProductLocatorGC(
                        scene=self._supported_scenes()[0],
                        origins=self._supported_origins()[0],
                        versions=version,
                    )
                    raise TestPassed()

    def test_init_unsupported_versions(self) -> None:
        for version in self._unsupported_versions():
            with self.subTest(version=version):
                with self.assertRaises(ValueError):
                    GridSatProductLocatorGC(
                        scene=self._supported_scenes()[0],
                        origins=self._supported_origins()[0],
                        versions=version,
                    )

    def test_get_base_url_supported_datasources(self) -> None:
        SUPPORTED_URLS: dict[str, str] = self._supported_urls()
        for datasource_id, expected_url in SUPPORTED_URLS.items():
            returned_url: str = self.locator.get_base_url(
                datasource=datasource_id
            )
            with self.subTest(datasource=datasource_id):
                self.assertEqual(returned_url, expected_url)

    def test_get_base_url_unsupported_datasources(self) -> None:
        for datasource_id in self._unsupported_datasources():
            with self.subTest(datasource=datasource_id):
                with self.assertRaises(ValueError):
                    self.locator.get_base_url(datasource=datasource_id)

    def test_get_datetime_valid_date_utc(self) -> None:
        # Expected `datetime` object set in UTC timezone.
        TZ_UTC: timezone = timezone.utc
        self._test_get_datetime_valid_date(12, TZ_UTC)

    def test_get_datetime_valid_date_non_utc(self) -> None:
        # Expected `datetime` object set in different non-UTC timezone.
        TZ_PYT: timezone = timezone(timedelta(hours=-4.0))
        self._test_get_datetime_valid_date(8, TZ_PYT)

    def test_get_datetime_bad_date_format(self) -> None:
        for invalid_filename in self._incomplete_dates():
            with self.subTest(filename=invalid_filename):
                with self.assertRaises(ValueError):
                    self.locator.get_datetime(invalid_filename)

    def test_get_datetime_incomplete_date(self) -> None:
        for invalid_filename in self._incomplete_dates():
            with self.subTest(filename=invalid_filename):
                with self.assertRaises(ValueError):
                    self.locator.get_datetime(invalid_filename)

    def test_get_datetime_invalid_date(self) -> None:
        # Valid filename with invalid date, e.g. September 31.
        VALID_FILENAME_INVALID_DATE: str = self._filename(month=9, day=31)
        with self.assertRaises(ValueError):
            self.locator.get_datetime(VALID_FILENAME_INVALID_DATE)

    def test_match_valid_filename(self) -> None:
        VALID_FILENAME: str = self._filename()
        self.assertTrue(self.locator.match(VALID_FILENAME))

    def test_match_misspelled_prefix(self) -> None:
        INVALID_FILENAME: str = self._filename(prefix="GridCat")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_supported_scenes(self) -> None:
        for scene in self._supported_scenes():
            self.setUpLocator(scene=scene)
            filename: str = self._filename(scene=scene)
            with self.subTest(scene=scene):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_scenes(self) -> None:
        self.setUpLocator(scene=self._supported_scenes()[0])
        for scene in self._unsupported_scenes():
            filename: str = self._filename(scene=scene)
            with self.subTest(scene=scene):
                self.assertFalse(self.locator.match(filename))

    def test_match_supported_origins(self) -> None:
        for origin in self._supported_origins():
            self.setUpLocator(origin=origin)
            filename: str = self._filename(origin=origin)
            with self.subTest(origin=origin):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_origins(self) -> None:
        self.setUpLocator(origin=self._supported_origins()[0])
        for origin in self._unsupported_origins():
            filename: str = self._filename(origin=origin)
            with self.subTest(origin=origin):
                self.assertFalse(self.locator.match(filename))

    def test_match_supported_multi_origins(self) -> None:
        for origin in self._supported_origins():
            filename: str = self._filename(origin=origin)
            with self.subTest(origin=origin):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_multi_origins(self) -> None:
        for origin in self._unsupported_origins():
            filename: str = self._filename(origin=origin)
            with self.subTest(origin=origin):
                self.assertFalse(self.locator.match(filename))

    def test_match_bad_date_format(self) -> None:
        for invalid_filename in self._bad_date_format():
            with self.subTest(filename=invalid_filename):
                self.assertFalse(self.locator.match(invalid_filename))

    def test_match_incomplete_date(self) -> None:
        for invalid_filename in self._incomplete_dates():
            with self.subTest(filename=invalid_filename):
                self.assertFalse(self.locator.match(invalid_filename))

    def test_match_invalid_date(self) -> None:
        # Valid filename format with invalid dates are not caught
        # by `match()` because the date format is correct and do not
        # violates the filename pattern, although it should be caught
        # by the `get_datetime()` method.
        #
        # See `test_get_datetime_invalid_date()`.
        VALID_FILENAME_INVALID_DATE: str = self._filename(month=9, day=31)
        self.assertTrue(self.locator.match(VALID_FILENAME_INVALID_DATE))

    def test_match_supported_versions(self) -> None:
        for version in self._supported_versions():
            self.setUpLocator(version=version)
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_versions(self) -> None:
        self.setUpLocator(version=self._supported_versions()[0])
        for version in self._unsupported_versions():
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertFalse(self.locator.match(filename))

    def test_match_supported_multi_versions(self) -> None:
        for version in self._supported_versions():
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_multi_versions(self) -> None:
        for version in self._unsupported_versions():
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertFalse(self.locator.match(filename))

    def test_match_wrong_suffix(self) -> None:
        INVALID_FILENAME: str = self._filename(suffix=".txt")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_scene_separator(self) -> None:
        INVALID_FILENAME: str = self._filename().replace("-GOES", "_GOES")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_origin_separator(self) -> None:
        INVALID_FILENAME: str = self._filename().replace(".goes12", "_goes12")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_date_separator(self) -> None:
        FILENAME: str = self._filename(year=1999, month=12, day=31, hour=23)
        invalid_filename: str = FILENAME.replace(".1999", "_1999")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Year separator"
        )
        invalid_filename = FILENAME.replace(".12", "_12")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Month separator"
        )
        invalid_filename = FILENAME.replace(".31", "_31")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Day separator"
        )
        invalid_filename = FILENAME.replace(".23", "_23")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Hour separator"
        )

    def test_match_bad_version_separator(self) -> None:
        INVALID_FILENAME: str = self._filename().replace(".v01", "_v01")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_suffix_separator(self) -> None:
        INVALID_FILENAME: str = self._filename(suffix="_nc")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_version_prefix(self) -> None:
        INVALID_FILENAME: str = self._filename().replace(".v01", ".b01")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_get_paths(self) -> None:
        TIME_1: datetime = datetime(1982, 1, 10, 11)
        TIME_2: datetime = datetime(1985, 12, 13, 14)
        for scene in self._supported_scenes():
            self.setUpLocator(scene=scene)
            returned_paths: list[str] = self.locator.get_paths(TIME_1, TIME_2)
            expected_paths: list[str] = [
                self._path(scene, year, month)
                for year in range(1982, 1986)
                for month in range(1, 13)
            ]
            with self.subTest(scene=scene):
                self.assertEqual(returned_paths, expected_paths)

    # ------------------------------------------------------------------

    def _test_get_datetime_valid_date(self, hour: int, tz: timezone) -> None:
        VALID_FILENAME: str = self._filename()
        expected_datetime: datetime = datetime(
            1994, 8, 23, hour, 13, tzinfo=tz
        )
        returned_datetime: datetime = self.locator.get_datetime(VALID_FILENAME)
        self.assertEqual(returned_datetime, expected_datetime)

    # ------------------------------------------------------------------

    @staticmethod
    def _filename(
        prefix: str = "GridSat",
        scene: str = "GOES",
        origin: str = "goes12",
        year: int = 1994,
        month: int = 8,
        day: int = 23,
        hour: int = 12,
        minute: int = 13,
        version: str = "v01",
        suffix: str = ".nc",
    ) -> str:
        # By default returns a valid filename for the GridSat-GOES/CONUS
        # imagery dataset products.
        SCENE_MAPPING: dict[str, str] = {
            "F": "GOES",
            "C": "CONUS",
            "M1": "MESOSCALE1",
            "M2": "MESOSCALE2",
        }
        if scene in SCENE_MAPPING:
            scene = SCENE_MAPPING[scene]
        ORIGIN_MAPPING: dict[str, str] = {
            f"G{id:02d}": f"goes{id:02d}" for id in range(8, 16)
        }
        if origin in ORIGIN_MAPPING:
            origin = ORIGIN_MAPPING[origin]
        return (
            f"{prefix}-{scene}.{origin}."
            f"{year:04d}.{month:02d}.{day:02d}.{hour:02d}{minute:02d}."
            f"{version}{suffix}"
        )

    @classmethod
    def _bad_date_format(cls) -> list[str]:
        BAD_YEAR: str = cls._filename(year=99999)
        BAD_MONTH: str = cls._filename(month=999)
        BAD_DAY: str = cls._filename(day=999)
        BAD_HOUR: str = cls._filename(hour=999)
        BAD_MINUTE: str = cls._filename(minute=999)
        return [BAD_YEAR, BAD_MONTH, BAD_DAY, BAD_HOUR, BAD_MINUTE]

    @classmethod
    def _incomplete_dates(cls) -> list[str]:
        NO_YEAR: str = cls._filename(year=9999).replace(".9999", "")
        NO_MONTH: str = cls._filename(month=99).replace(".99", "")
        NO_DAY: str = cls._filename(day=99).replace(".99", "")
        NO_HOUR: str = cls._filename(hour=99).replace(".99", "")
        NO_MINUTE: str = cls._filename(minute=99).replace("99", "")
        return [NO_YEAR, NO_MONTH, NO_DAY, NO_HOUR, NO_MINUTE]

    @staticmethod
    def _path(scene: str, year: int, month: int) -> str:
        SCENE_MAPPING: dict[str, str] = {"F": "goes", "C": "conus"}
        return f"{SCENE_MAPPING[scene]}/{year:04d}/{month:02d}/"

    # ------------------------------------------------------------------

    @staticmethod
    def _available_datasources() -> dict[str, str]:
        return {
            "AWS": "s3://noaa-cdr-gridsat-goes-pds/data/",
            "Azure": "https://noaa-cdr-gridsat-goes.blob.core.windows.net/"
            "data/",
            "GCP": "gs://noaa-cdr-gridsat-goes/data/",
            "NOAA": "https://www.ncei.noaa.gov/data/gridsat-goes/access/",
        }

    @classmethod
    def _supported_urls(cls) -> dict[str, str]:
        AVAILABLE_DATASOURCES: dict[str, str] = cls._available_datasources()
        SUPPORTED_DATASOURCES: list[str] = cls._supported_datasources()
        return {
            datasource_id: url
            for datasource_id, url in AVAILABLE_DATASOURCES.items()
            if datasource_id in SUPPORTED_DATASOURCES
        }

    @staticmethod
    def _supported_datasources() -> list[str]:
        return ["NOAA"]

    @classmethod
    def _unsupported_datasources(cls) -> list[str]:
        AVAILABLE_DATASOURCES: dict[str, str] = cls._available_datasources()
        SUPPORTED_DATASOURCES: list[str] = cls._supported_datasources()
        return [
            datasource_id
            for datasource_id in AVAILABLE_DATASOURCES.keys()
            if datasource_id not in SUPPORTED_DATASOURCES
        ]

    # ------------------------------------------------------------------

    @staticmethod
    def _supported_scenes() -> list[str]:
        return ["F", "C"]

    @staticmethod
    def _unsupported_scenes() -> list[str]:
        return ["M1", "M2"]

    # ------------------------------------------------------------------

    @staticmethod
    def _get_origins(a: int, b: int) -> list[str]:
        return [f"G{n:02d}" for n in range(a, b + 1)]

    @classmethod
    def _supported_origins(cls) -> list[str]:
        return cls._get_origins(8, 15)

    @classmethod
    def _unsupported_origins(cls) -> list[str]:
        AVAILABLE_ORIGINS: list[str] = cls._get_origins(1, 18)
        SUPPORTED_ORIGINS: list[str] = cls._supported_origins()
        return [
            origin
            for origin in AVAILABLE_ORIGINS
            if origin not in SUPPORTED_ORIGINS
        ]

    # ------------------------------------------------------------------

    @staticmethod
    def _supported_versions() -> list[str]:
        return ["v01"]

    @classmethod
    def _unsupported_versions(cls) -> list[str]:
        AVAILABLE_VERSIONS: set[str] = {f"v{v:02}" for v in range(1, 10)}
        SUPPORTED_VERSIONS: list[str] = cls._supported_versions()
        return [
            version
            for version in AVAILABLE_VERSIONS
            if version not in SUPPORTED_VERSIONS
        ]
